"""Tokenize tank definition."""
import wtrwrks.waterworks.tank as ta
import wtrwrks.tanks.utils as ut
import wtrwrks.string_manipulations.diff as di
import numpy as np


class Tokenize(ta.Tank):
  """The defintion of the Tokenize tank. Contains the implementations of the _pour and _pump methods, as well as which slots and tubes the waterwork objects will look for.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """

  func_name = 'tokenize'
  slot_keys = ['strings', 'tokenizer', 'detokenizer', 'max_len']
  tube_keys = ['target', 'tokenizer', 'detokenizer', 'diff']
  pass_through_keys = ['tokenizer', 'detokenizer']

  def _pour(self, strings, tokenizer, max_len, detokenizer):
    """Execute the Tokenize tank (operation) in the pour (forward) direction.

    Parameters
    ----------
    strings: np.ndarray of strings
      The array of strings to tokenize.
    tokenizer: func
      Function which converts a string into a list of strings.
    detokenizer: func
      Function which takens in a list of tokens and returns a string. Not strictly necessary but it makes the tube 'diff' much smaller if it's close to the real method of detokenizing.
    max_len: int
      The maximum number of tokens. Defines the size of the added dimension.

    Returns
    -------
    dict(
      target: np.ndarray
        The array of tokenized strings. Will have rank = rank('a') + 1 where the last dimesion will have size max_len.
      tokenizer: func
        Function which converts a string into a list of strings.
      detokenizer: func
        Function which takens in a list of tokens and returns a string. Not strictly necessary but it makes the tube 'diff' much smaller if it's close to the real method of detokenizing.
      diff: np.ndarray of strings
        The array of strings which define the differences between the original string and the string that has been tokenized then detokenized.
    )

    """
    # Convert to a numpy array.
    strings = np.array(strings)

    # print detokenizer('I went on a run yesterday. I saw a bird and it was magnificent. I hope to see one again tomorrow.')
    # Handle the empty array case
    if not strings.size:
      return {'target': ut.maybe_copy(strings), 'diff': ut.maybe_copy(strings), 'tokenizer': tokenizer, 'detokenizer': detokenizer}

    all_tokens = []
    all_diffs = []

    lengths = []
    for string in strings.flatten():
      # Tokenize the string, and regularize the length of the array by padding
      # with '' to fill out the array if it's too small or truncated if it's
      # too long.
      tokens = np.array(tokenizer(string))
      lengths.append(len(tokens))
      if tokens.size < max_len:
        num = max_len - tokens.size
        tokens = np.concatenate([tokens, np.full([num], '')])
      else:
        tokens = tokens[:max_len]

      all_tokens.append(tokens)
      # Detokenize the tokens and reconstruct the orignal string from the
      # diff_string
      processed = detokenizer(tokens)
      diff = di.get_diff_string(processed, string)
      all_diffs.append(np.array(diff, dtype=np.unicode))

    # Combine all the tokens arrays into a single array and reshape to the
    # shape of the original strings array with an additional dimesion of size
    # max_len.
    token_array = np.stack(all_tokens)
    target = np.reshape(token_array, list(strings.shape) + [max_len])

    # Keep all the string diffs and reshape it to match the original strings
    # array shape.
    diff_array = np.stack(all_diffs)
    diff = np.reshape(diff_array, strings.shape)

    return {'target': target, 'diff': diff, 'tokenizer': tokenizer, 'detokenizer': detokenizer}

  def _pump(self, target, diff, tokenizer, detokenizer):
    """Execute the Tokenize tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    target: np.ndarray
      The array of tokenized strings. Will have rank = rank('a') + 1 where the last dimesion will have size max_len.
    tokenizer: func
      Function which converts a string into a list of strings.
    detokenizer: func
      Function which takens in a list of tokens and returns a string. Not strictly necessary but it makes the tube 'diff' much smaller if it's close to the real method of detokenizing.
    diff: np.ndarray of strings
      The array of strings which define the differences between the original string and the string that has been tokenized then detokenized.

    Returns
    -------
    dict(
      strings: np.ndarray of strings
        The array of strings to tokenize.
      tokenizer: func
        Function which converts a string into a list of strings.
      detokenizer: func
        Function which takens in a list of tokens and returns a string. Not strictly necessary but it makes the tube 'diff' much smaller if it's close to the real method of detokenizing.
      max_len: int
        The maximum number of tokens. Defines the size of the added dimension.
    )

    """
    max_len = target.shape[-1]
    all_strings = []

    # Flatten out all the dimensions aside from the token dimesion of target and
    # flatten out the diff array completely to iterate through them.
    for tokens, diff_string in zip(np.reshape(target, (-1, max_len)), diff.flatten()):

      # Detokenize and reconstruct the original string from the diff string.
      string = detokenizer(tokens)
      string = di.reconstruct(string, diff_string)
      all_strings.append(string)
    all_strings = np.array(all_strings)

    # Reshape to the original shape.
    strings = np.reshape(all_strings, target.shape[:-1])
    return {'strings': strings, 'tokenizer': tokenizer, 'max_len': max_len, 'detokenizer': detokenizer}


class MultiTokenize(ta.Tank):
  """The defintion of the Tokenize tank. Contains the implementations of the _pour and _pump methods, as well as which slots and tubes the waterwork objects will look for.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """

  func_name = 'multi_tokenize'
  slot_keys = ['strings', 'selector', 'tokenizers', 'detokenizers', 'max_len']
  tube_keys = ['target', 'selector', 'tokenizers', 'detokenizers', 'diff']
  pass_through_keys = ['tokenizers', 'detokenizers', 'selector']

  def _pour(self, strings, selector, tokenizers, max_len, detokenizers):
    """Execute the Tokenize tank (operation) in the pour (forward) direction.

    Parameters
    ----------
    strings: np.ndarray of strings
      The array of strings to tokenize.
    tokenizer: func
      Function which converts a string into a list of strings.
    detokenizer: func
      Function which takens in a list of tokens and returns a string. Not strictly necessary but it makes the tube 'diff' much smaller if it's close to the real method of detokenizing.
    max_len: int
      The maximum number of tokens. Defines the size of the added dimension.

    Returns
    -------
    dict(
      target: np.ndarray
        The array of tokenized strings. Will have rank = rank('a') + 1 where the last dimesion will have size max_len.
      tokenizer: func
        Function which converts a string into a list of strings.
      detokenizer: func
        Function which takens in a list of tokens and returns a string. Not strictly necessary but it makes the tube 'diff' much smaller if it's close to the real method of detokenizing.
      diff: np.ndarray of strings
        The array of strings which define the differences between the original string and the string that has been tokenized then detokenized.
    )

    """
    # Convert to a numpy array.
    strings = np.array(strings)
    selector = np.array(selector)

    if strings.shape != selector.shape:
      raise ValueError("Shape of strings and selector must match. Got {} and {}".format(strings.shape, selector.shape))

    # print detokenizer('I went on a run yesterday. I saw a bird and it was magnificent. I hope to see one again tomorrow.')
    # Handle the empty array case
    if not strings.size:
      return {'target': ut.maybe_copy(strings), 'selector': ut.maybe_copy(selector), 'diff': ut.maybe_copy(strings), 'tokenizers': tokenizers, 'detokenizers': detokenizers}

    all_tokens = []
    all_diffs = []

    lengths = []
    for string, language in zip(strings.flatten(), selector.flatten()):
      # Tokenize the string, and regularize the length of the array by padding
      # with '' to fill out the array if it's too small or truncated if it's
      # too long.
      tokenizer = tokenizers[language]
      tokens = np.array(tokenizer(string))
      lengths.append(len(tokens))
      if tokens.size < max_len:
        num = max_len - tokens.size
        tokens = np.concatenate([tokens, np.full([num], '')])
      else:
        tokens = tokens[:max_len]

      all_tokens.append(tokens)
      # Detokenize the tokens and reconstruct the orignal string from the
      # diff_string
      detokenizer = detokenizers[language]
      processed = detokenizer(tokens)
      diff = di.get_diff_string(processed, string)
      all_diffs.append(np.array(diff, dtype=np.unicode))

    # Combine all the tokens arrays into a single array and reshape to the
    # shape of the original strings array with an additional dimesion of size
    # max_len.
    token_array = np.stack(all_tokens)
    target = np.reshape(token_array, list(strings.shape) + [max_len])

    # Keep all the string diffs and reshape it to match the original strings
    # array shape.
    diff_array = np.stack(all_diffs)
    diff = np.reshape(diff_array, strings.shape)

    return {'target': target, 'diff': diff, 'tokenizers': tokenizers, 'detokenizers': detokenizers, 'selector': selector}

  def _pump(self, target, selector, diff, tokenizers, detokenizers):
    """Execute the Tokenize tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    target: np.ndarray
      The array of tokenized strings. Will have rank = rank('a') + 1 where the last dimesion will have size max_len.
    tokenizer: func
      Function which converts a string into a list of strings.
    detokenizer: func
      Function which takens in a list of tokens and returns a string. Not strictly necessary but it makes the tube 'diff' much smaller if it's close to the real method of detokenizing.
    diff: np.ndarray of strings
      The array of strings which define the differences between the original string and the string that has been tokenized then detokenized.

    Returns
    -------
    dict(
      strings: np.ndarray of strings
        The array of strings to tokenize.
      tokenizer: func
        Function which converts a string into a list of strings.
      detokenizer: func
        Function which takens in a list of tokens and returns a string. Not strictly necessary but it makes the tube 'diff' much smaller if it's close to the real method of detokenizing.
      max_len: int
        The maximum number of tokens. Defines the size of the added dimension.
    )

    """
    max_len = target.shape[-1]
    all_strings = []

    # Flatten out all the dimensions aside from the token dimesion of target and
    # flatten out the diff array completely to iterate through them.
    for tokens, language, diff_string in zip(np.reshape(target, (-1, max_len)), selector.flatten(), diff.flatten()):

      # Detokenize and reconstruct the original string from the diff string.
      detokenizer = detokenizers[language]
      string = detokenizer(tokens)
      string = di.reconstruct(string, diff_string)
      all_strings.append(string)
    all_strings = np.array(all_strings)

    # Reshape to the original shape.
    strings = np.reshape(all_strings, target.shape[:-1])
    return {'strings': strings, 'selector': selector,  'tokenizers': tokenizers, 'max_len': max_len, 'detokenizers': detokenizers}



import unittest
import wtrwrks.utils.test_helpers as th
import wtrwrks.tanks.tank_defs as td
import numpy as np


class TestSplit(th.TestTank):

  def test_one_d(self):
    self.pour_pump(
      td.split,
      {
        'a': np.array([1, 2, 3]),
        'indices': np.array([1, 2]),
        'axis': 0
      },
      {
        'target': [np.array([1]), np.array([2]), np.array([3])],
        'indices': np.array([1, 2]),
        'axis': 0
      },
      type_dict={'a': np.ndarray, 'indices': np.ndarray, 'axis': int}
    )
    self.pour_pump(
      td.split,
      {
        'a': np.array([1, 2, 3]),
        'indices': np.array(3),
        'axis': 0
      },
      {
        'target': [np.array([1]), np.array([2]), np.array([3])],
        'indices': np.array(3),
        'axis': 0
      },
      type_dict={'a': np.ndarray, 'indices': np.ndarray, 'axis': int}
    )

  def test_two_d(self):
    a = np.array([[0, 1], [2, 3], [4, 5], [1, 0]])
    self.pour_pump(
      td.split,
      {
        'a': a,
        'indices': np.array([2, 4]),
        'axis': 0
      },
      {
        'target': np.split(a, np.array([2, 4])),
        'indices': np.array([2, 4]),
        'axis': 0
      },
      type_dict={'a': np.ndarray, 'indices': np.ndarray, 'axis': int}
    )
    a = np.array([[0, 1], [2, 3], [4, 5], [1, 0]])
    self.pour_pump(
      td.split,
      {
        'a': a,
        'indices': [1, 2],
        'axis': 1
      },
      {
        'target': np.split(a, [1, 2], axis=1),
        'indices': [1, 2],
        'axis': 1
      },
      type_dict={'a': np.ndarray, 'indices': np.ndarray, 'axis': int}
    )


if __name__ == "__main__":
    unittest.main()



"""FlatTokenize tank definition."""
import wtrwrks.waterworks.tank as ta
import wtrwrks.tanks.utils as ut
import wtrwrks.string_manipulations.diff as di
import numpy as np
# from chop.mmseg import Tokenizer as MMSEGTokenizer
# from chop.hmm import Tokenizer as HMMTokenizer
import nltk
import tinysegmenter


class FlatTokenize(ta.Tank):
  """The defintion of the FlatTokenize tank. Contains the implementations of the _pour and _pump methods, as well as which slots and tubes the waterwork objects will look for.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """
  func_name = 'flat_tokenize'
  slot_keys = ['strings', 'tokenizer', 'detokenizer', 'ids']
  tube_keys = ['target', 'tokenizer', 'detokenizer', 'diff', 'shape', 'ids']
  pass_through_keys = ['tokenizer', 'detokenizer', 'ids']

  def _pour(self, strings, ids, tokenizer, detokenizer=lambda a: ' '.join(a)):
    """Execute the FlatTokenize tank (operation) in the pour (forward) direction.

    Parameters
    ----------
    strings: np.ndarray of strings
      The array of strings to tokenize.
    tokenizer: func
      Function which converts a string into a list of strings.
    detokenizer: func
      Function which takens in a list of tokens and returns a string. Not strictly necessary but it makes the tube 'diff' much smaller if it's close to the real method of detokenizing.
    ids: np.ndarray
      An array of ids which uniquely identify each element of 'strings'. Necessary in order to reconstruct strings since all information about axis is lost when flattened. Each id from ids must be unique.The array of is the same shape as strings

    Returns
    -------
    dict(
      target: np.ndarray
        A one dimensional array of tokens.
      tokenizer: func
        Function which converts a string into a list of strings.
      detokenizer: func
        Function which takens in a list of tokens and returns a string. Not strictly necessary but it makes the tube 'diff' much smaller if it's close to the real method of detokenizing.
      diff: np.ndarray of strings
        The array of strings which define the differences between the original string and the string that has been tokenized then detokenized.
      shape: list of ints
        The shape of the inputted array.
      ids: np.ndarray
        An array of ids which uniquely identify each element of 'strings'. Necessary in order to reconstruct strings. The array of is the same shape as target
    )

    """
    strings = np.array(strings)
    # Guard for the empty array case
    if not strings.size:
      return {'target': ut.maybe_copy(strings), 'diff': ut.maybe_copy(strings), 'tokenizer': tokenizer, 'detokenizer': detokenizer}

    all_tokens = []
    all_diffs = []

    # Go through each element of the string array, and it's corresponding id.
    r_ids = []
    for string_id, string in zip(ids.flatten(), strings.flatten()):
      # Tokenize the string and add it to the long list of all the tokens.
      tokens = tokenizer(string)
      all_tokens.extend(tokens)

      # Copy the string id len(tokens) times so that the ids always have
      # the same length as the tokens. This makes it more suitable for breaking
      # up in downstream tanks.
      r_ids.extend([string_id] * len(tokens))

      # Find the string diff after detokenizing the tokens.
      processed = detokenizer(tokens)
      diff = di.get_diff_string(processed, string)

      # Copy the diff len(tokens) times so that it always has the same size
      # as tokens. This makes it more suitable for breaking up in downstream
      # tanks.
      all_diffs.extend([diff] * len(tokens))

    target = np.array(all_tokens).astype(strings.dtype)
    diff = np.array(all_diffs).astype(strings.dtype)
    r_ids = np.array(r_ids)

    return {'target': target, 'diff': diff, 'tokenizer': tokenizer, 'detokenizer': detokenizer, 'ids': r_ids, 'shape': strings.shape}

  def _pump(self, target, diff, tokenizer, detokenizer, ids, shape):
    """Execute the FlatTokenize tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    target: np.ndarray
      A one dimensional array of tokens.
    tokenizer: func
      Function which converts a string into a list of strings.
    detokenizer: func
      Function which takens in a list of tokens and returns a string. Not strictly necessary but it makes the tube 'diff' much smaller if it's close to the real method of detokenizing.
    diff: np.ndarray of strings
      The array of strings which define the differences between the original string and the string that has been tokenized then detokenized.
    shape: list of ints
      The shape of the inputted array.
    ids: np.ndarray
      An array of ids which uniquely identify each element of 'strings'. Necessary in order to reconstruct strings. The array of is the same shape as target

    Returns
    -------
    dict(
      strings: np.ndarray of strings
        The array of strings to tokenize.
      tokenizer: func
        Function which converts a string into a list of strings.
      detokenizer: func
        Function which takens in a list of tokens and returns a string. Not strictly necessary but it makes the tube 'diff' much smaller if it's close to the real method of detokenizing.
      ids: np.ndarray
        An array of ids which uniquely identify each element of 'strings'. Necessary in order to reconstruct strings since all information about axis is lost when flattened. Each id from ids must be unique.The array of is the same shape as strings
    )

    """
    # Handle the empty array case
    if not ids.size:
      return {'strings': target.reshape([0] + list(shape[1:])), 'tokenizer': tokenizer, 'detokenizer': detokenizer}

    # Get all the unique ids, and the indices which point to that unique value
    # value. Sort them, so that the indices are pulled out in sequential order.
    unique_ids, indices = np.unique(ids, return_index=True)
    indices = np.sort(indices)

    all_strings = []
    r_ids = []
    for index in indices:

      # Pull out all the tokens belonging to a particular id.
      string_id = ids[index]
      mask = (ids == string_id)
      tokens = target[mask].flatten()

      # Pull out the corresponding diff_string. Since it was duplicated many
      # times we only need to pull out the first one.
      diff_string = diff[mask][0]

      # Detokenize the tokens and reconstruct the orignal string from the
      # diff_string

      string = detokenizer(tokens)
      string = di.reconstruct(string, diff_string)

      r_ids.append(string_id)
      all_strings.append(string)
    r_ids = np.reshape(r_ids, shape)
    strings = np.reshape(all_strings, shape)
    return {'strings': strings, 'tokenizer': tokenizer, 'detokenizer': detokenizer, 'ids': r_ids}



import json
import numpy as np
import pandas as pd
import dill as pickle
import os
import subprocess as sp
import tensorflow as tf

def untar_dir(file_to_untar, dir_to_save_to, verbose=False):
  """Untar a tarball and save it to another directory

  Parameters
  ----------
  file_to_untar : str
    The file name of the tarball
  dir_to_save_to : str
      The directory to save the untarred directory

  """
  if verbose:
    tar_str = "tar zxvf "
  else:
    tar_str = "tar zxf "
  command = tar_str + file_to_untar + " -C " + dir_to_save_to
  stdout = sp.call(command, shell=True, stdout=sp.PIPE, stderr=sp.PIPE)


def tar_dir(dir_to_tar, dir_to_save_to, ignore_patterns=None, verbose=False):
  """Tar up a directory and save to a particular location.

  Parameters
  ----------
  dir_to_tar : str
    Directory to tar up
  dir_to_save_to : str
    The directory to save the tarred directory

  """
  tar_file_name = dir_to_tar.rstrip('/').split('/')[-1] + '.tar.gz'
  tar_file_name = os.path.join(dir_to_save_to, tar_file_name)

  if verbose:
    tar_str = "tar zcvf "
  else:
    tar_str = "tar zcf "

  if ignore_patterns is None:
    command = tar_str + tar_file_name + " " + dir_to_tar
  else:
    excludes = ' '
    for pat in ignore_patterns:
      excludes += "--exclude='" + pat + "' "
    command = tar_str + tar_file_name + excludes + dir_to_tar

  stdout = sp.call(command, shell=True, stdout=sp.PIPE)


def create_dirs(dir):
  """Create directories and all intermediary directories if they don't exist.

  Parameters
  ----------
  dir : str
    Name of the directory

  """
  if not os.path.exists(dir):
    os.makedirs(dir)


def maybe_create_dir(*args):
  """Create directories and all intermediary directories if they don't exist.

  Parameters
  ----------
  *args:
      The path of the directories to maybe create.

  Returns
  -------
  str
    The new directory path
  """
  if not args:
    return './'

  full_dir = os.path.join(*args)
  if not os.path.isdir(full_dir):
      os.makedirs(full_dir)
  return full_dir


def save_to_file(obj, file_name):
  """Wrapper to automatically find the proper way to save a file

  Parameters
  ----------
  obj :
    Object to save to file
  file_name : str
    File name to save the file to.


  """
  # Create the directory path
  dir = file_name.split('/')[:-1]
  if dir:
    maybe_create_dir(*dir)

  # Find file type and then save using appropriate function.
  file_type = file_name.split('.')[-1]
  with open(file_name, 'w') as obj_file:
    if file_type == 'json':
      json.dump(obj, obj_file)
    elif file_type == 'npy':
      np.save(obj_file, obj)
    elif file_type == 'pickle':
      pickle.dump(obj, obj_file)
  if file_type == 'h5':
    store = pd.HDFStore(file_name)
    store['df'] = obj


def read_from_file(file_name):
  """Wrapper to automatically find the proper way to read a file

  Parameters
  ----------
  file_name : str
    File name of file to read

  Returns
  -------
  obj :
    Object that was read from file
  """

  # Find file type from the extension and then read using appropriate
  # function.
  file_type = file_name.split('.')[-1]
  with open(file_name, 'r') as obj_file:
    if file_type == 'json':
      obj = json.load(obj_file)
    elif file_type == 'npy':
      obj = np.load(obj_file)
    elif file_type == 'pickle':
      obj = pickle.load(obj_file)
  if file_type == 'h5':
    store = pd.HDFStore(file_name)
    obj = store['df']

  return obj



"""OneHot tank definition."""
import wtrwrks.waterworks.waterwork_part as wp
import wtrwrks.waterworks.tank as ta
import wtrwrks.tanks.utils as ut
import numpy as np


class OneHot(ta.Tank):
  """The defintion of the OneHot tank. Contains the implementations of the _pour and _pump methods, as well as which slots and tubes the waterwork objects will look for.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """

  func_name = 'one_hot'
  slot_keys = ['indices', 'depth']
  tube_keys = ['target', 'missing_vals']

  def _pour(self, indices, depth):
    """Execute the OneHot tank (operation) in the pour (forward) direction.

    Parameters
    ----------
    indices: np.ndarray of ints
      The array of indices to be one hotted.
    depth: int
      The maximum allowed index value and the size of the n + 1 dimension of the outputted array.

    Returns
    -------
    dict(
      target: np.ndarray
        The one hotted array.
      missing_vals: list of ints
        The indices which were not in the range of 0 <= i < depth
    )

    """
    indices = np.array(indices)
    if not indices.shape:
      target = np.zeros([depth], dtype=np.float64)
      missing_vals = -2
      if indices < 0 or indices >= depth:
        missing_vals = indices
      else:
        target[indices] = 1.0
      return {'target': target, 'missing_vals': missing_vals}
    # Pull out all the indices which are not in the range 0 <= index <
    # depth.
    mask = (indices < 0) | (indices >= depth)

    missing_vals = np.ones(indices.shape, dtype=int) * -2
    missing_vals[mask] = indices[mask]
    # Reshape the indices to give them a new dimension. Compare them to
    # each index between 0 and depth - 1, find the location where they are
    # True, and turn the Trues into 1's
    indices = np.expand_dims(indices, axis=-1)
    target = (np.arange(depth) == indices).astype(np.float64)

    return {'target': target, 'missing_vals': missing_vals}

  def _pump(self, target, missing_vals):
    """Execute the OneHot tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    target: np.ndarray
      The one hotted array.
    missing_vals: list of ints
      The indices which were not in the range of 0 <= i < depth

    Returns
    -------
    dict(
      indices: np.ndarray of ints
        The array of indices to be one hotted.
      depth: int
        The maximum allowed index value and the size of the n + 1 dimension of the outputted array.
    )

    """
    if len(target.shape) == 1:
      indices = missing_vals if not np.where(target > 0)[0] else np.where(target > 0)[0][0]

      return {'indices': indices, 'depth': target.shape[0]}
    # Start the indices all as -1.
    indices = -1 * np.ones(target.shape[:-1], missing_vals.dtype)
    depth = int(target.shape[-1])

    # If the target is a one dimensional array, then just set to either the
    # missing val in the case that it's a zero hot or to the index of the
    # non zero value.

    # If the target is more than on dimensional, then first find all the
    # locations where there are non-zeros. Use those locations to build an
    # array of shape target.shape[:-1] and use the last dimension of the
    # 'where' array to set the the index. Replace all the -1's with the
    # missing vals.
    unpacked_indices = np.where(target > 0)
    locs = unpacked_indices[:-1]
    vals = unpacked_indices[-1]

    indices[locs] = vals
    mask = indices == -1
    indices[mask] = missing_vals[mask]
    return {'indices': indices, 'depth': depth}



"""HalfWidth tank definition."""
import wtrwrks.waterworks.tank as ta
import wtrwrks.tanks.utils as ut
import wtrwrks.string_manipulations.diff as di
import numpy as np
# from chop.mmseg import Tokenizer as MMSEGTokenizer
# from chop.hmm import Tokenizer as HMMTokenizer
import unicodedata


def _half_width(string):
  return unicodedata.normalize('NFKC', unicode(string))


class HalfWidth(ta.Tank):
  """The defintion of the HalfWidth tank. Contains the implementations of the _pour and _pump methods, as well as which slots and tubes the waterwork objects will look for.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """
  func_name = 'half_width'
  slot_keys = ['strings']
  tube_keys = ['target', 'diff']

  def _pour(self, strings):
    """Execute the HalfWidth tank (operation) in the pour (forward) direction.

    Parameters
    ----------
    strings: np.ndarray of unicode
      The array of unicode characters to be converted to half width

    Returns
    -------
    dict(
      target: np.ndarray of unicode
        The array of half widthed strings.
      diff: np.darray of unicode
        The string difference between the original strings and the half widthed strings.
    )

    """
    strings = np.array(strings)
    target = np.vectorize(_half_width)(strings)

    diff = np.vectorize(di.get_diff_string)(target, strings)

    return {'target': target, 'diff': diff}

  def _pump(self, target, diff):
    """Execute the HalfWidth tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    target: np.ndarray of unicode
      The array of half widthed strings.
    diff: np.darray of unicode
      The string difference between the original strings and the half widthed strings.

    Returns
    -------
    dict(
      strings: np.ndarray of unicode
        The array of unicode characters to be converted to half width
    )

    """

    strings = np.vectorize(di.reconstruct)(target, diff)

    return {'strings': strings}



# -*- coding: utf-8 -*-
import unittest
import wtrwrks.utils.test_helpers as th
import wtrwrks.tanks.tank_defs as td
import numpy as np
import tinysegmenter

class TestLowerCase(th.TestTank):
  def test_scalar(self):
    strings = "It is what it is."
    diff = '[["i", 0, 1, "I"]]'
    self.pour_pump(
      td.lower_case,
      {
        'strings': strings,
      },
      {
        'target': strings.lower(),
        'diff': diff,
      },
      test_type=False
    )

  def test_one_d(self):
    strings = np.array([
      u'チラシ・勧誘印刷物の無断投函は一切お断り',
      u'すみませんが、もう一度どお願いします。'
    ])
    diff = ['[]', '[]']
    self.pour_pump(
      td.lower_case,
      {
        'strings': strings,
      },
      {
        'target': strings,
        'diff': diff,
      },
      test_type=False
    )

  def test_two_d(self):
    strings = np.array([
      [
        "It is what it is.",
        "Whatever, Bob's mother has seen the world."
      ],
      [
        "The sun is not yellow, it's chicken. OK.",
        "Don't need a weatherman to know which way the wind blows."
      ]
    ])
    diff = [
      ['[["i", 0, 1, "I"]]', '[["i", 0, 1, "W"], ["i", 10, 11, "B"]]'],
      ['[["i", 0, 1, "T"], ["i", 37, 39, "OK"]]', '[["i", 0, 1, "D"]]']
    ]

    self.pour_pump(
      td.lower_case,
      {
        'strings': strings,
      },
      {
        'target': np.reshape([s.lower() for s in strings.flatten()], strings.shape),
        'diff': diff,
      },
      test_type=False
    )

if __name__ == "__main__":
    unittest.main()



"""Tube definition."""
import wtrwrks.waterworks.waterwork_part as wp
from wtrwrks.waterworks.empty import Empty, empty
import os

class Tube(wp.WaterworkPart):
  """Object that is always part of some tank which stores the output (in the pour or forward direction) of the operation perfomed by the tank and connects to the slot of another tank.

  Attributes
  ----------
  tank : Tank
    The tank that this tube is a part of.
  key : str
    The string to identify the tube within the tank. Must be unique among all other tubes of this tank.
  val : some data type or None
    The value last outputed by the tank (i.e. operation), if applicable.
  slot : Slot or None
    The slot from the other tank this tube is connected to, if applicable.
  name : str
    The string used to identify the tube within the entire waterwork. Must be unique among all other tubes of this waterwork.

  """

  def __init__(self, tank, key, val=None, slot=empty, plug=None, downstream_tube=None):
    """Initialize the tube.

    Parameters
    ----------
    tank : Tank
      The tank that this tube is a part of.
    key : str
      The string to identify the tube within the tank. Must be unique among all other tubes of this tank.
    val : some data type or None
      The value last outputed by the tank (i.e. operation), if applicable.
    slot : Slot or None
      The slot from the other tank this tube is connected to, if applicable.
    """
    self.key = key
    self.tank = tank
    self.slot = slot
    self.name = None
    self.val = val
    self.plug = None
    self.downstream_tube = downstream_tube

    super(Tube, self).__init__(tank.waterwork, self.name)
    if self.name in self.waterwork.tubes:
      raise ValueError(self.name + " already defined as tube. Choose a different name.")
    if plug is not None:
      self.set_plug(plug)

  def __add__(self, other):
    """Define an add tank (operation) between two tubes."""
    import wtrwrks.tanks.tank_defs as td
    return td.add(a=self, b=other, waterwork=self.waterwork)

  def __div__(self, other):
    """Define an add tank (operation) between two tubes."""
    import wtrwrks.tanks.tank_defs as td
    return td.div(a=self, b=other, waterwork=self.waterwork)

  def __eq__(self, other):
    """Determine whether two tubes are the same within one waterwork."""
    return (self.tank, self.key) == (other.tank, other.key)

  def __hash__(self):
    """Determine whether two tubes are the same within one waterwork."""
    return hash((self.tank, self.key))

  def __mul__(self, other):
    """Define an add tank (operation) between two tubes."""
    import wtrwrks.tanks.tank_defs as td
    return td.mul(a=self, b=other, waterwork=self.waterwork)

  def __radd__(self, other):
    """Define an add tank (operation) between two tubes."""
    import wtrwrks.tanks.tank_defs as td
    return td.add(a=other, b=self, waterwork=self.waterwork)

  def __rdiv__(self, other):
    """Define an add tank (operation) between two tubes."""
    import wtrwrks.tanks.tank_defs as td
    return td.div(a=other, b=self, waterwork=self.waterwork)

  def __rmul__(self, other):
    """Define an add tank (operation) between two tubes."""
    import wtrwrks.tanks.tank_defs as td
    return td.mul(a=other, b=self, waterwork=self.waterwork)

  def __rsub__(self, other):
    """Define an add tank (operation) between two tubes."""
    import wtrwrks.tanks.tank_defs as td
    return td.sub(a=other, b=self, waterwork=self.waterwork)

  def __str__(self):
    """Get a string of the name of the tube."""
    return self.name

  def __sub__(self, other):
    """Define an add tank (operation) between two tubes."""
    import wtrwrks.tanks.tank_defs as td
    return td.sub(a=self, b=other, waterwork=self.waterwork)

  def _get_default_name(self, prefix=''):
    """Set a default name. Must be defined by subclass."""
    return os.path.join(self.tank.name, 'tubes', self.key)

  def _save_dict(self):
    save_dict = {}
    save_dict['key'] = self.key
    save_dict['slot'] = None if type(self.slot) is Empty else self.slot.name
    save_dict['name'] = self.name
    save_dict['tank'] = self.tank.name
    save_dict['plug'] = self.plug
    save_dict['downstream_tube'] = None if self.downstream_tube is None else self.downstream_tube.name

    return save_dict

  def get_tuple(self):
    """Get a tuple that describes the tube."""
    return (self.tank.name, self.key)

  def get_val(self):
    """Get the value stored in the tube."""
    return self.val

  def set_val(self, val):
    """Set the value stored in the tube."""
    self.val = val

  def set_plug(self, plug, downstream=True):
    """Plug up the tap with a function."""
    if self.downstream_tube is not None and downstream:
      self = self.downstream_tube

    if self.name not in self.waterwork.taps:
      raise ValueError("Can only plug taps. " + str(self) + " is not a tap.")

    if not callable(plug):
      func_plug = lambda a: plug
    else:
      func_plug = plug
    self.plug = func_plug

  def set_name(self, name, downstream=True):
    """Set the name of the tube within the waterwork."""
    # If it has a downstream tube, then actually set that name. Unless
    # downstream is set to false
    tu = self
    if self.downstream_tube is not None and downstream:
      tu = self.downstream_tube

    old_name = tu.name
    tu.name = name

    full_name_space = tu.name_space._get_name_string()
    if full_name_space:
      full_name_space = full_name_space + '/'

    if type(name) not in (str, unicode):
      raise TypeError("'name' must be of type str or unicode. Got " + str(type(name)))
    elif not tu.name.startswith(full_name_space):
      tu.name = os.path.join(tu.name_space._get_name_string(), tu.name)

    if tu.name in tu.waterwork.tubes:
      raise ValueError(tu.name + " already defined as tube. Choose a different name.")

    del tu.waterwork.tubes[old_name]
    tu.waterwork.tubes[tu.name] = tu

    if old_name in tu.waterwork.taps:
      del tu.waterwork.taps[old_name]
      tu.waterwork.taps[tu.name] = tu

  def unplug(self, downstream=True):
    """Remove the plug function from a tap."""
    if self.downstream_tube is not None and downstream:
      self = self.downstream_tube
    self.plug = None



"""Partition tank definition."""
import wtrwrks.waterworks.waterwork_part as wp
import wtrwrks.waterworks.tank as ta
import wtrwrks.tanks.utils as ut
import numpy as np

class Partition(ta.Tank):
  """The defintion of the Partition tank. Contains the implementations of the _pour and _pump methods, as well as which slots and tubes the waterwork objects will look for.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """

  func_name = 'partition'
  slot_keys = ['a', 'ranges']
  tube_keys = ['target', 'ranges', 'missing_cols', 'missing_array']
  pass_through_keys = ['ranges']

  def _pour(self, a, ranges):
    """Execute the Partition tank (operation) in the pour (forward) direction.

    Parameters
    ----------
    a: np.ndarray
      The array to take slices from.
    ranges: np.ndarray
      The index ranges of each slice. The first dimension can be of any size and the second dimension must be two.

    Returns
    -------
    dict(
      target: list of arrays
        The list of array slices. The length of the list is equal to the size of the first dimension of 'ranges'.
      ranges: np.ndarray of ints
        The index ranges of each slice. The first dimension can be of any size and the second dimension must be two.
      missing_cols: np.ndarray of ints
        The columns of the array that were not selected by the slices defined by 'ranges'
      missing_array: np.ndarray
        The slices of array that were not selected by the slices defined by 'ranges'.
    )

    """
    # Cast everything to a numpy array.
    a = np.array(a)
    ranges = np.array(ranges)
    full_cols = np.arange(a.shape[0], dtype=int)

    target = []
    all_ranges = []
    # Go through each col range which defines a slice.
    for col_range in ranges:
      target.append(a[col_range[0]: col_range[1]])

      # Add all the column numbers in the slice to an array.
      all_ranges.append(np.arange(col_range[0], col_range[1]))

    # Find all the columns which did not appear in some slice. Save those
    # columns and extract the corresponding slices from the array 'a'
    all_ranges = np.concatenate(all_ranges, axis=0)
    missing_cols = np.setdiff1d(full_cols, all_ranges)
    missing_array = a[missing_cols]

    return {'target': target, 'ranges': ranges, 'missing_cols': missing_cols, 'missing_array': missing_array}

  def _pump(self, target, ranges, missing_cols, missing_array):
    """Execute the Partition tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    target: list of arrays
      The list of array slices. The length of the list is equal to the size of the first dimension of 'ranges'.
    ranges: np.ndarray of ints
      The index ranges of each slice. The first dimension can be of any size and the second dimension must be two.
    missing_cols: np.ndarray of ints
      The columns of the array that were not selected by the slices defined by 'ranges'
    missing_array: np.ndarray
      The slices of array that were not selected by the slices defined by 'ranges'.

    Returns
    -------
    dict(
      a: np.ndarray
        The array to take slices from.
      ranges: np.ndarray
        The index ranges of each slice. The first dimension can be of any size and the second dimension must be two.
    )

    """
    # Get the total length of the partitioned dimension. Handling the empty
    # array case.
    ranges = np.array(ranges)
    if target or missing_cols:
      max_index = np.max(np.concatenate([ranges[:, 1] - 1, missing_cols.flatten()]))
    else:
      max_index = -1

    # If the target list isn't empty then take the sizes of the dimensions not
    # involved in the partition. Otherwise, use the missing array to get the
    # shape.
    if target:
      inner_dims = target[0].shape[1:]
    else:
      inner_dims = missing_array.shape[1:]

    # Create the empty array
    a = np.zeros([max_index + 1] + list(inner_dims), dtype=missing_array.dtype)

    # Go through each of the partitions to fill the array a back up with
    # the slices.
    for subarray, col_range in zip(target, ranges):
      a[col_range[0]: col_range[1]] = subarray

    # Add any of the missing columns back in as well.
    for col_num, col in enumerate(missing_cols):
      a[col] = missing_array[col_num]

    return {'a': a, 'ranges': ranges}


class PartitionByIndex(ta.Tank):
  """The defintion of the Partition tank. Contains the implementations of the _pour and _pump methods, as well as which slots and tubes the waterwork objects will look for.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """

  func_name = 'partition_by_index'
  slot_keys = ['a', 'indices']
  tube_keys = ['target', 'indices', 'missing_cols', 'missing_array']
  pass_through_keys = ['indices']

  def _pour(self, a, indices):
    """Execute the Partition tank (operation) in the pour (forward) direction.

    Parameters
    ----------
    a: np.ndarray
      The array to take slices from.
    indices: np.ndarray
      The indices of all the slices.

    Returns
    -------
    dict(
      target: list of arrays
        The list of array slices. The length of the list is equal to the size of the first dimension of 'indices'.
      indices: np.ndarray of ints
        The indices of all the slices.
      missing_cols: np.ndarray of ints
        The columns of the array that were not selected by the slices defined by 'indices'
      missing_array: np.ndarray
        The slices of array that were not selected by the slices defined by 'indices'.
    )

    """
    # Cast everything to a numpy array.
    a = np.array(a)
    full_cols = np.arange(a.shape[0], dtype=int)

    target = []
    all_used_indices = []
    # Go through each col range which defines a slice.
    for cols in indices:
      target.append(a[list(cols)])

      # Add all the column numbers in the slice to an array.
      all_used_indices.append(cols)

    # Find all the columns which did not appear in some slice. Save those
    # columns and extract the corresponding slices from the array 'a'
    all_used_indices = np.concatenate(all_used_indices, axis=0)
    missing_cols = np.setdiff1d(full_cols, all_used_indices)
    missing_array = a[missing_cols]

    return {'target': target, 'indices': indices, 'missing_cols': missing_cols, 'missing_array': missing_array}

  def _pump(self, target, indices, missing_cols, missing_array):
    """Execute the Partition tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    target: list of arrays
      The list of array slices. The length of the list is equal to the size of the first dimension of 'indices'.
    indices: np.ndarray of ints
      The indices of all the slices.
    missing_cols: np.ndarray of ints
      The columns of the array that were not selected by the slices defined by 'indices'
    missing_array: np.ndarray
      The slices of array that were not selected by the slices defined by 'indices'.

    Returns
    -------
    dict(
      a: np.ndarray
        The array to take slices from.
      indices: np.ndarray
        The indices of all the slices.
    )

    """
    # Get the total length of the partitioned dimension. Handling the empty
    # array case.
    if target or missing_cols:
      all_indices = np.concatenate(indices, axis=0)
      max_index = np.max(np.concatenate([all_indices, missing_cols.flatten()], axis=0))
    else:
      max_index = -1

    # If the target list isn't empty then take the sizes of the dimensions not
    # involved in the partition. Otherwise, use the missing array to get the
    # shape.
    if target:
      inner_dims = target[0].shape[1:]
    else:
      inner_dims = missing_array.shape[1:]

    # Create the empty array
    a = np.zeros([max_index + 1] + list(inner_dims), dtype=missing_array.dtype)

    # Go through each of the partitions to fill the array a back up with
    # the slices.
    for subarray, cols in zip(target, indices):
      a[list(cols)] = subarray

    # Add any of the missing columns back in as well.
    for col_num, col in enumerate(missing_cols):
      a[col] = missing_array[col_num]

    return {'a': a, 'indices': indices}



import unittest
import wtrwrks.utils.test_helpers as th
# import wtrwrks.tanks.add as ad
import wtrwrks.tanks.tank_defs as td
import numpy as np


class TestAdd(th.TestTank):
  def test_int(self):
    self.pour_pump(td.add, {'a': 1, 'b': 2}, {'target': 3, 'smaller_size_array': 2, 'a_is_smaller': False})

  def test_array_int(self):
    self.pour_pump(
      td.add,
      {'a': np.array([1, 2]), 'b': 2},
      {'target': np.array([3, 4]), 'smaller_size_array': 2, 'a_is_smaller': False},
      type_dict={'a': np.ndarray, 'b': int}
    )

  def test_int_array(self):
    self.pour_pump(
      td.add,
      {'a': 2, 'b': np.array([1, 2])},
      {'target': np.array([3, 4]), 'smaller_size_array': 2, 'a_is_smaller': True},
      type_dict={'a': int, 'b': np.ndarray}
    )

  def test_array(self):
    self.pour_pump(
      td.add,
      {'a': np.array([1, 2]), 'b': np.array([3, 4])},
      {'target': np.array([4, 6]), 'smaller_size_array': np.array([3, 4]), 'a_is_smaller': False},
      type_dict={'a': np.ndarray, 'b': np.ndarray}
    )

    self.pour_pump(
      td.add,
      {'a': np.array([2]), 'b': np.array([3, 4])},
      {'target': np.array([5, 6]), 'smaller_size_array': np.array([2]), 'a_is_smaller': True},
      type_dict={'a': np.ndarray, 'b': np.ndarray}
    )

    self.pour_pump(
      td.add,
      {'b': np.array([2]), 'a': np.array([3, 4])},
      {'target': np.array([5, 6]), 'smaller_size_array': np.array([2]), 'a_is_smaller': False},
      type_dict={'a': np.ndarray, 'b': np.ndarray}
    )

if __name__ == "__main__":
    unittest.main()



"""Empty definition."""
import wtrwrks.waterworks.waterwork_part as wp


class Empty(wp.WaterworkPart):
  """A special type of waterwork part, used to start waterworks without having the data at definition time.

  Attributes
  ----------
  tank : Tank
    The tank that this tube is a part of.
  key : str
    The string to identify the tube within the tank. Must be unique among all other tubes of this tank.
  val : some data type or None
    The value last outputed by the tank (i.e. operation), if applicable.
  slot : Slot or None
    The slot from the other tank this tube is connected to, if applicable.
  name : str
    The string used to identify the tube within the entire waterwork. Must be unique among all other tubes of this waterwork.

  """

  def __init__(self):
    """Initialize the tube.

    Parameters
    ----------
    tank : Tank
      The tank that this tube is a part of.
    key : str
      The string to identify the tube within the tank. Must be unique among all other tubes of this tank.
    val : some data type or None
      The value last outputed by the tank (i.e. operation), if applicable.
    slot : Slot or None
      The slot from the other tank this tube is connected to, if applicable.

    """

  def __add__(self, other):
    """Define an add tank (operation) between two tubes."""
    import wtrwrks.tanks.tank_defs as td
    return td.add(a=self, b=other)

  def __div__(self, other):
    """Define an add tank (operation) between two tubes."""
    import wtrwrks.tanks.tank_defs as td
    return td.div(a=self, b=other)

  def __eq__(self, other):
    """Determine whether two tubes are the same within one waterwork."""
    return type(self) is Empty and type(other) is Empty

  def __ne__(self, other):
    """Determine whether two tubes are the same within one waterwork."""
    return not (self == other)

  def __mul__(self, other):
    """Define an add tank (operation) between two tubes."""
    import wtrwrks.tanks.tank_defs as td
    return td.mul(a=self, b=other)

  def __radd__(self, other):
    """Define an add tank (operation) between two tubes."""
    import wtrwrks.tanks.tank_defs as td
    return td.add(a=other, b=self)

  def __rdiv__(self, other):
    """Define an add tank (operation) between two tubes."""
    import wtrwrks.tanks.tank_defs as td
    return td.div(a=other, b=self)

  def __rmul__(self, other):
    """Define an add tank (operation) between two tubes."""
    import wtrwrks.tanks.tank_defs as td
    return td.mul(a=other, b=self)

  def __rsub__(self, other):
    """Define an add tank (operation) between two tubes."""
    import wtrwrks.tanks.tank_defs as td
    return td.sub(a=other, b=self)

  def __str__(self):
    """Get a string of the name of the tube."""
    return "EmptySlot"

  def __sub__(self, other):
    """Define an add tank (operation) between two tubes."""
    import wtrwrks.tanks.tank_defs as td
    return td.sub(a=self, b=other)


empty = Empty()



"""Print tank definition."""
import wtrwrks.waterworks.waterwork_part as wp
import wtrwrks.waterworks.tank as ta
import numpy as np
import wtrwrks.tanks.utils as ut

class Print(ta.Tank):
  """Get the shape of an array.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """

  func_name = 'print_val'
  slot_keys = None
  tube_keys = None
  pass_through_keys = None

  def _pour(self, **kwargs):
    """

    Parameters
    ----------
    kwargs:
      The arguments to print.

    Returns
    -------
    dict(
      kwargs:
        The original arguments
    )

    """
    for key in kwargs:
      print kwargs[key]
    return kwargs

  def _pump(self, **kwargs):
    """Execute the Shape tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    kwargs:
      The orignal arguments.

    Returns
    -------
    dict(
      kwargs:
        The aruments to print.
    )

    """
    for key in kwargs:
      print kwargs[key]
    return kwargs



"""Clone tank definition."""
import wtrwrks.waterworks.tank as ta
import wtrwrks.tanks.utils as ut
import numpy as np
from wtrwrks.waterworks.empty import empty


class DoNothing(ta.Tank):
  """Dummy tank only used if you want to make something part of the waterwork but don't want to actually do anything to it.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """
  func_name = 'do_nothing'
  slot_keys = ['a']
  tube_keys = ['target']

  def _pour(self, a):
    """Execute the DoNothing tank (operation) in the pour (forward) direction.

    Parameters
    ----------
    a: object
      The object to be added to the waterwork.

    Returns
    -------
    dict(
      target: object
        The object to be added to the waterwork.
    )

    """
    return {'target': a}

  def _pump(self, target):
    """Execute the Clone tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    target: object
      The object to be added to the waterwork.

    Returns
    -------
    dict(
      a: object
        The object to be added to the waterwork.
    )

    """
    return {'a': target}


class Clone(ta.Tank):
  """The defintion of the Clone tank. Contains the implementations of the _pour and _pump methods, as well as which slots and tubes the waterwork objects will look for.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """
  func_name = 'clone'
  slot_keys = ['a']
  tube_keys = ['a', 'b']

  def _pour(self, a):
    """Execute the Clone tank (operation) in the pour (forward) direction.

    Parameters
    ----------
    a: object
      The object to be cloned into two.

    Returns
    -------
    dict(
      a: type of slot 'a'
        The first of the two cloned objects.
      b: type of slot 'a'
        The second of the two cloned objects.
    )

    """
    return {'a': ut.maybe_copy(a), 'b': ut.maybe_copy(a)}

  def _pump(self, a, b):
    """Execute the Clone tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    a: type of slot 'a'
      The first of the two cloned objects.
    b: type of slot 'a'
      The second of the two cloned objects.

    Returns
    -------
    dict(
      a: object
        The object to be cloned into two.
    )

    """
    return {'a': ut.maybe_copy(a)}


class CloneMany(ta.Tank):
  """Clone an object many times (rather than just once).

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """
  func_name = 'merge_equal'
  slot_keys = ['a', 'num']
  tube_keys = None
  test_equal = None

  def _pour(self, a):
    """Execute the Clone tank (operation) in the pour (forward) direction.

    Parameters
    ----------
    a: object
      The object to be cloned into two.
    num: int > 0
      The number of clones

    Returns
    -------
    dict(
      a0: object
        zeroth clone
      a1: object
        first clone,
      .
      .
      .
    )

    """

    r_dict = {}
    for key in self.tube_keys:
      r_dict[key] = ut.maybe_copy(a)
    return r_dict

  def _pump(self, **kwargs):
    """Execute the Clone tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    a0: object
      zeroth clone
    a1: object
      first clone,
    .
    .
    .

    Returns
    -------
    dict(
      a: object
        The object to be cloned into two.
    )

    """
    r_dict = {}
    for key in self.slot_keys:
      r_dict['a'] = ut.maybe_copy(kwargs[key])
      break
    return r_dict


class MergeEqual(ta.Tank):
  """Merge several equal objects into a single object. All tubes must have equal values, otherwise you will get unexpected results. (Opposite of clone_many)

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """
  func_name = 'merge_equal'
  slot_keys = None
  tube_keys = ['target']
  test_equal = None

  def _pour(self, **kwargs):
    """Execute the MergeEqual tank (operation) in the pour (forward) direction.

    Parameters
    ----------
    a0: object
      zeroth equal object
    a1: object
      first equal object,
    .
    .
    .

    Returns
    -------
    dict(
      a: type of slot 'a'target: object
        The merged object. Simply takes the value of the first in the list.
    )

    """
    if self.test_equal:
      for key in kwargs:
        if not np.all(kwargs[key] == kwargs['a0']):
          raise ValueError("All arguments passed to merge_equal must be equal. Got " + str(kwargs[key]) + ' and ' + str(kwargs['a0']))
    return {'target': ut.maybe_copy(kwargs['a0'])}

  def _pump(self, target):
    """Execute the Clone tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    target: object
      The merged object. Simply takes the value of the first in the list.

    Returns
    -------
    dict(
      a0: object
        zeroth equal object
      a1: object
        first equal object,
      .
      .
      .
    )

    """
    kwargs = {}
    for key in self.slot_keys:
      kwargs[key] = ut.maybe_copy(target)
    return kwargs

  def _save_dict(self):
    save_dict = {}
    save_dict['func_name'] = self.func_name
    save_dict['name'] = self.name
    save_dict['args'] = [empty for _ in self.slot_keys]

    return save_dict



"""DatetimeToNum tank definition."""
import wtrwrks.waterworks.waterwork_part as wp
import wtrwrks.waterworks.tank as ta
import numpy as np
import datetime


class DatetimeToNum(ta.Tank):
  """The defintion of the DatetimeToNum tank. Contains the implementations of the _pour and _pump methods, as well as which slots and tubes the waterwork objects will look for.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """

  func_name = 'datetime_to_num'
  slot_keys = ['a', 'zero_datetime', 'num_units', 'time_unit']
  tube_keys = ['target', 'zero_datetime', 'num_units', 'time_unit', 'diff']
  pass_through_keys = ['num_units', 'time_unit', 'zero_datetime']

  def _pour(self, a, zero_datetime, num_units, time_unit):
    """Execute the DatetimeToNum tank (operation) in the pour (forward) direction.

    Parameters
    ----------
    a: np.ndarray of datetime64
      The array of datetimes to be converted to numbers.
    zero_datetime: datetime64
      The datetime that will be considered zero when converted to a number. All other datetimes will be relative to the this.
    num_units: int
      This along with 'time_unit' define the time resolution to convert to numbers. e.g. if 'time_unit' is 'D' (day) and 'num_units' is 2 then every two days corresponds to 1 increment in time.
    time_unit: str - from 'Y', 'M', 'W', 'D', 'h', 'm', 's', 'ms', 'us', 'ns', 'ps', 'fs', 'as'
      This along with 'num_units' define the time resolution to convert to numbers. e.g. if 'time_unit' is 'D' (day) and 'num_units' is 2 then every two days corresponds to 1 increment in time.

    Returns
    -------
    dict(
      target: np.ndarray of
        The array of datetimes that were converted to numbers.
      zero_datetime: datetime64
        The datetime that will be considered zero when converted to a number. All other datetimes will be relative to the this.
      num_units: int
        This along with 'time_unit' define the time resolution to convert to numbers. e.g. if 'time_unit' is 'D' (day) and 'num_units' is 2 then every two days corresponds to 1 increment in time.
      time_unit: str - from 'Y', 'M', 'W', 'D', 'h', 'm', 's', 'ms', 'us', 'ns', 'ps', 'fs', 'as'
        This along with 'num_units' define the time resolution to convert to numbers. e.g. if 'time_unit' is 'D' (day) and 'num_units' is 2 then every two days corresponds to 1 increment in time.
      diff: np.ndarray of timedelta64
        The difference between the original array 'a' and the array which lost information from taking everything according to a finite time resolution. e.g. For num_units=1 and time_unit='M' datetime(2000, 3, 4) gets mapped to datetime(2000, 3) so the diff would be a timedelta64 of 4 days.
    )

    """
    a = np.array(a, dtype=np.datetime64)
    zero_datetime = np.array(zero_datetime, dtype=np.datetime64)
    target = (a - zero_datetime)/np.timedelta64(num_units, time_unit)

    # Save the diff since information is lost depending on the size of time unit
    undone = target * np.timedelta64(num_units, time_unit) + zero_datetime
    diff = a - undone

    # If there is no diff then don't bother saving any information.
    # if np.array(diff == np.array(0, dtype='timedelta64[us]')).all():
    #   diff = np.array([], dtype='timedelta64[us]')

    return {'target': target, 'zero_datetime': zero_datetime, 'num_units': num_units, 'time_unit': time_unit, 'diff': diff}

  def _pump(self, target, zero_datetime, num_units, time_unit, diff):
    """Execute the DatetimeToNum tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    target: np.ndarray of
      The array of datetimes that were converted to numbers.
    zero_datetime: datetime64
      The datetime that will be considered zero when converted to a number. All other datetimes will be relative to the this.
    num_units: int
      This along with 'time_unit' define the time resolution to convert to numbers. e.g. if 'time_unit' is 'D' (day) and 'num_units' is 2 then every two days corresponds to 1 increment in time.
    time_unit: str - from 'Y', 'M', 'W', 'D', 'h', 'm', 's', 'ms', 'us', 'ns', 'ps', 'fs', 'as'
      This along with 'num_units' define the time resolution to convert to numbers. e.g. if 'time_unit' is 'D' (day) and 'num_units' is 2 then every two days corresponds to 1 increment in time.
    diff: np.ndarray of timedelta64
      The difference between the original array 'a' and the array which lost information from taking everything according to a finite time resolution. e.g. For num_units=1 and time_unit='M' datetime(2000, 3, 4) gets mapped to datetime(2000, 3) so the diff would be a timedelta64 of 4 days.

    Returns
    -------
    dict(
      a: np.ndarray of datetime64
        The array of datetimes to be converted to numbers.
      zero_datetime: datetime64
        The datetime that will be considered zero when converted to a number. All other datetimes will be relative to the this.
      num_units: int
        This along with 'time_unit' define the time resolution to convert to numbers. e.g. if 'time_unit' is 'D' (day) and 'num_units' is 2 then every two days corresponds to 1 increment in time.
      time_unit: str - from 'Y', 'M', 'W', 'D', 'h', 'm', 's', 'ms', 'us', 'ns', 'ps', 'fs', 'as'
        This along with 'num_units' define the time resolution to convert to numbers. e.g. if 'time_unit' is 'D' (day) and 'num_units' is 2 then every two days corresponds to 1 increment in time.
    )

    """

    # Get back the original multiplying by the time resolution and adding the
    # zero_date plus any diff values.
    a = target * np.timedelta64(num_units, time_unit) + zero_datetime
    if diff.size:
       a = a + diff

    return {'a': a, 'zero_datetime': zero_datetime, 'num_units': num_units, 'time_unit': time_unit}



# -*- coding: utf-8 -*-
import unittest
import wtrwrks.utils.test_helpers as th
import wtrwrks.tanks.tank_defs as td
import numpy as np
import tinysegmenter

class TestTokenize(th.TestTank):
  def test_scalar(self):
    tokenizer = lambda s: s.split()
    detokenizer = lambda a: ' '.join(a)
    strings = "It is what it is."
    diff = '[["d", 17, 22, ""]]'
    self.pour_pump(
      td.tokenize,
      {
        'strings': strings,
        'tokenizer': tokenizer,
        'max_len': 10,
        'detokenizer': detokenizer
      },
      {
        'target': ['It', 'is', 'what', 'it', 'is.', '', '', '', '', ''],
        'diff': diff,
        'tokenizer': tokenizer,
        'detokenizer': detokenizer
      },
      test_type=False
    )

  def test_one_d(self):
    ts = tinysegmenter.TinySegmenter()
    tokenizer = ts.tokenize
    detokenizer = lambda a: ''.join(a)
    strings = np.array([
      u'チラシ・勧誘印刷物の無断投函は一切お断り',
      u'すみませんが、もう一度どお願いします。'
    ])
    diff = ['[["i", 16, 16, "\\u5207\\u304a\\u65ad\\u308a"]]', '[["i", 18, 18, "\\u3002"]]']
    target = [[u'\u30c1\u30e9\u30b7', u'\u30fb', u'\u52e7\u8a98', u'\u5370\u5237', u'\u7269', u'\u306e', u'\u7121\u65ad', u'\u6295\u51fd', u'\u306f', u'\u4e00'], [u'\u3059\u307f\u307e\u305b', u'\u3093', u'\u304c', u'\u3001', u'\u3082\u3046', u'\u4e00', u'\u5ea6', u'\u3069\u304a\u9858\u3044', u'\u3057', u'\u307e\u3059']]
    self.pour_pump(
      td.tokenize,
      {
        'strings': strings,
        'tokenizer': tokenizer,
        'max_len': 10,
        'detokenizer': detokenizer
      },
      {
        'target': target,
        'diff': diff,
        'tokenizer': tokenizer,
        'detokenizer': detokenizer
      },
      test_type=False
    )

  def test_two_d(self):
    tokenizer = lambda s: s.split()
    detokenizer = lambda a: ' '.join(a)
    strings = np.array([
      [
        "It is what it is.",
        "Whatever, Bob's mother has seen the world."
      ],
      [
        "The sun is not yellow, it's chicken. OK.",
        "Don't need a weatherman to know which way the wind blows."
      ]
    ])
    diff = [
      ['[["d", 17, 22, ""]]', '[["d", 42, 45, ""]]'],
      ['[["d", 40, 42, ""]]', '[["i", 50, 50, " blows."]]']
    ]

    target = np.array(
      [
        [
          ['It', 'is', 'what', 'it', 'is.', '', '', '', '', ''], ['Whatever,', "Bob's", 'mother', 'has', 'seen', 'the', 'world.', '', '', '']
        ],
        [
          ['The', 'sun', 'is', 'not', 'yellow,', "it's", 'chicken.', 'OK.', '', ''],
          ["Don't", 'need', 'a', 'weatherman', 'to', 'know', 'which', 'way', 'the', 'wind']
        ]
      ])
    self.pour_pump(
      td.tokenize,
      {
        'strings': strings,
        'tokenizer': tokenizer,
        'max_len': 10,
        'detokenizer': detokenizer
      },
      {
        'target': target,
        'diff': diff,
        'tokenizer': tokenizer,
        'detokenizer': detokenizer
      },
      test_type=False
    )

if __name__ == "__main__":
    unittest.main()



import numpy as np
import tensorflow as tf
import os
import wtrwrks.utils.dir_functions as d
import imp
import inspect as i
import datetime
import unittest
import shutil
import tempfile
import wtrwrks.waterworks.waterwork as wa
import pprint
import pandas as pd


class WWTest(unittest.TestCase):
  def setUp(self):
      self.temp_dir = tempfile.mkdtemp()

  def tearDown(self):
      shutil.rmtree(self.temp_dir)

  def equals(self, first, second, test_type=True):

    if type(first) in (np.dtype, type(int)):
      pass
    elif type(first) is list:
      self.assertEqual(len(first), len(second))
      for f, s in zip(first, second):
        try:
          if type(f) is np.ndarray:
            self._arrays_equal(f, s)
          elif type(f) is float and np.isnan(f):
            self.assertTrue(np.isnan(s))
          else:
            self.assertEqual(f, s)
        except(ValueError, AttributeError, AssertionError) as e:
          print "FIRST", f, type(f)
          print "SECOND", s, type(s)
          raise e
    elif type(first) is not np.ndarray:
      try:
        self.assertEqual(first, second)
        if test_type:
          self.assertTrue(type(first) is type(second))
      except (ValueError, AssertionError) as e:
        print "FIRST", first, type(first)
        print "SECOND", second, type(second)
        raise e
    else:
      self._arrays_equal(first, second, test_type=test_type)

  def _arrays_equal(self, first, second, test_type=True):
    try:
      self.assertTrue(arrays_equal(first, second, test_type=test_type))
    except AssertionError as e:
      print 'SHAPES', first.shape, np.array(second).shape
      print "FIRST ", first.tolist(), first.dtype
      print "SECOND", second, np.array(second).dtype
      if first.shape == np.array(second).shape:
        mask = first == second
        print 'DIFF', first[~mask], second[~mask]
      raise e


class TestTank (WWTest):
  def pour_pump(self, tank_func, input_dict, output_dict, type_dict=None, test_type=True):
    with wa.Waterwork() as ww:
      # test eager
      tank_tubes, tank_slots = tank_func(**input_dict)

      for key in input_dict:
        tank_slots[key].unplug()

      tank = tank_tubes[tank_tubes.keys()[0]].tank
      out_dict = {t: v.val for t, v in tank_tubes.iteritems()}
      self.assertEqual(sorted(out_dict.keys()), sorted(output_dict.keys()))
      for key in out_dict:
        try:
          self.equals(out_dict[key], output_dict[key], test_type=test_type)
        # except (ValueError, AssertionError) as e:
        except (AssertionError) as e:
          print 'Pour direction, key:', key
          raise e

    # test pour
    out_dict = tank.pour(**input_dict)
    out_dict = {t: v for t, v in out_dict.iteritems()}

    self.assertEqual(sorted(out_dict.keys()), sorted(output_dict.keys()))
    for key in out_dict:
      try:
        self.equals(out_dict[key], output_dict[key], test_type=test_type)
      except (ValueError, AssertionError) as e:
        print 'Pour direction, key:', key
        raise e

    in_dict = tank.pump(**out_dict)

    self.assertEqual(sorted(in_dict.keys()), sorted(input_dict.keys()))
    for key in in_dict:
      try:
        self.equals(in_dict[key], input_dict[key], test_type=test_type)
      except (ValueError, AssertionError) as e:
        print 'Pump direction, key:', key
        raise e


class TestTransform(WWTest):
  def pour_pump(self, trans, array, output_dict, test_type=True):
    # trans.calc_global_values(array)
    tap_dict = trans.pour(array)
    out_dict = {str(k): v for k, v in tap_dict.iteritems()}

    self.assertEqual(sorted(out_dict.keys()), sorted(output_dict.keys()))
    for key in out_dict:
      try:
        self.equals(out_dict[key], output_dict[key], test_type=test_type)
      except (ValueError, AssertionError) as e:
        print 'Pour direction, key:', key
        raise e

    original = trans.pump(out_dict)
    try:
      self.equals(original, array, test_type=test_type)
    except (ValueError, AssertionError) as e:
      raise e

  def write_read(self, trans, temp_dir):
    temp_file_path = os.path.join(temp_dir, 'temp.pickle')
    trans.save_to_file(temp_file_path)
    cls = trans.__class__
    trans = cls(from_file=temp_file_path)

    return trans

  def write_read_example(self, trans, array, dir, test_type=True, num_cols=None):
    remade_array = None
    tap_dict = trans.pour(array)
    example_dicts = trans.tap_dict_to_examples(tap_dict)
    file_name = os.path.join(dir, 'temp.tfrecord')
    writer = tf.python_io.TFRecordWriter(file_name)

    for feature_dict in example_dicts:
      example = tf.train.Example(
        features=tf.train.Features(feature=feature_dict)
      )
      writer.write(example.SerializeToString())

    writer.close()

    dataset = tf.data.TFRecordDataset(file_name)

    if num_cols is None:
      num_cols = array.shape[1]

    dataset = dataset.map(trans.read_and_decode)
    iter = tf.data.Iterator.from_structure(
      dataset.output_types,
      dataset.output_shapes
    )
    init = iter.make_initializer(dataset)
    features = iter.get_next()

    for key in features:
      features[key] = tf.reshape(features[key], [-1])
    with tf.Session() as sess:
      sess.run(init)

      example_dicts = []
      try:
        while True:
          example_dict = sess.run(features)
          example_dicts.append(example_dict)
      except tf.errors.OutOfRangeError:
        pass
    tap_dict = trans.examples_to_tap_dict(example_dicts)
    remade_array = trans.pump(tap_dict)
    if type(array) is pd.DataFrame:
      array = array.values
    self.equals(array, remade_array, test_type)

class TestDataset (TestTransform):
  def pour_pump(self, dt, array, output_dict, test_type=True):
    # dt.calc_global_values(array)
    tap_dict = dt.pour(array)

    out_dict = {str(k): v for k, v in tap_dict.iteritems()}

    self.assertEqual(sorted(out_dict.keys()), sorted(output_dict.keys()))
    for key in out_dict:
      try:
        self.equals(out_dict[key], output_dict[key], test_type=test_type)
      except (ValueError, AssertionError) as e:
        raise e

    original = dt.pump(out_dict)

    try:
      self.equals(original, array, test_type=test_type)
    except (ValueError, AssertionError) as e:
      raise e


def arrays_equal(first, second, threshold=0.001, test_type=True):
  first = np.array(first, copy=True)
  second = np.array(second, copy=True)

  # Check that the arrays are the same shape.
  if first.shape != second.shape:
    return False
  if test_type and first.dtype != second.dtype:
    return False
  if first.size == 0 and second.size == 0:
    return True

  if np.issubdtype(first.dtype, np.datetime64) or np.issubdtype(first.dtype, np.timedelta64):
    if not (np.isnat(first) == np.isnat(second)).all():
      return False
    if np.issubdtype(first.dtype, np.datetime64):
      default = datetime.datetime(1970, 1, 1)
    else:
      default = datetime.timedelta(0)

    first[np.isnat(first)] = default
    second[np.isnat(second)] = default

    return (first == second).all()
  elif not np.issubdtype(first.dtype, np.number) and not np.issubdtype(second.dtype, np.number):
    # print '-'*20
    # mask = first.astype(np.unicode) != second.astype(np.unicode)
    # print first[mask], second[mask]
    # print '-'*20
    return (first.astype(np.unicode) == second.astype(np.unicode)).all()
  try:
    if not (np.isnan(first.astype(np.float64)) == np.isnan(second.astype(np.float64))).all():
      return False

    if not (np.isneginf(first.astype(np.float64)) == np.isneginf(second.astype(np.float64))).all():
      return False

    if not (np.isposinf(first.astype(np.float64)) == np.isposinf(second.astype(np.float64))).all():
      return False

    first[np.isnan(first.astype(np.float64))] = 0.0
    second[np.isnan(second.astype(np.float64))] = 0.0

    first[np.isneginf(first.astype(np.float64))] = 0.0
    second[np.isneginf(second.astype(np.float64))] = 0.0

    first[np.isposinf(first.astype(np.float64))] = 0.0
    second[np.isposinf(second.astype(np.float64))] = 0.0
  except ValueError:
    pass
  if threshold is None or not np.issubdtype(first.dtype, np.number):
    return (first == second).all()
  else:
    return (np.abs(first - second) < threshold).all()


def assert_arrays_equal(test_obj, a1, a2, threshold=None):
  """Shortcut for testing whether two numpy arrays are close enough to equal. Prints out the arrays and throws a unittest exception if they aren't.

  Parameters
  ----------
  test_obj : unittest subclass instance
    The unittest subclass instance to send the exceptions to.
  a1 : np.array
    First array
  a2 : np.array
    Second array
  threshold : float
    The allowed difference in values in a1 and a2 to still be considered equal

  """
  a1 = np.array(a1, copy=True)
  a2 = np.array(a2, copy=True)
  # Check that the arrays are the same shape.
  test_obj.assertEqual(a1.shape, a2.shape)

  if np.issubdtype(a1.dtype, np.datetime64) or np.issubdtype(a1.dtype, np.timedelta64):
    test_obj.assertTrue(
      (np.isnat(a1) == np.isnat(a2)).all()
    )
    a1[np.isnat(a1)] = datetime.datetime(1970, 1, 1)
    a2[np.isnat(a2)] = datetime.datetime(1970, 1, 1)

  try:
    test_obj.assertTrue(
      (np.isnan(a1.astype(np.float64)) == np.isnan(a2.astype(np.float64))).all()
    )
    a1[np.isnan(a1.astype(np.float64))] = 0.0
    a2[np.isnan(a2.astype(np.float64))] = 0.0
  except ValueError:
    pass

  # Catch the exception so the arrays can be printed first.
  try:
    # If no threshold is given then require exact match. Otherwise, test if
    # the values are just within some threshold of each other.

    if threshold is None:
      test_obj.assertTrue((a1 == a2).all())
    else:
      test_obj.assertTrue((np.abs(a1 - a2) < threshold).all())

  except Exception as e:
    # Print the arrays then raise the orginal exception.
    print a1
    print a2
    print a1.dtype
    raise e


def assert_tensor_equal_array(test_obj, a1, a2, threshold=None):
  """Shortcut for testing whether the evaled tensor is equal to some array. It will create the session, etc. so that does not need to be done outside this function.

  Parameters
  ----------
  test_obj : unittest subclass instance
    The unittest subclass instance to send the exceptions to.
  a1 : tf.tensor
    The tensor to be evaled then tested against the second array
  a2 : np.array
    Second array
  threshold : float
    The allowed difference in values in a1 and a2 to still be considered equal

  """
  # Initialize any variables that may exist in the graph
  init_op = tf.group(
    tf.global_variables_initializer(),
    tf.local_variables_initializer()
  )

  # Run the session to eval the tensor.
  with tf.Session() as sess:
    sess.run(init_op)
    a = sess.run(a1)

  # Check that the arrays are the same shape.
  test_obj.assertEqual(a.shape, a2.shape)

  # Catch the exception so the arrays can be printed first.
  try:
    # If no threshold is given then require exact match. Otherwise, test if
    # the values are just within some threshold of each other.
    if threshold is None:
      test_obj.assertTrue((a == a2).all())
    else:
      test_obj.assertTrue((np.abs(a - a2) < threshold).all())

  except Exception as e:
    # Print the arrays then raise the orginal exception.
    print a
    print a2
    raise e


def print_tensor(a1, to_list=False):
  """Shortcut for testing whether the evaled tensor is equal to some array. It will create the session, etc. so that does not need to be done outside this function.

  Parameters
  ----------
  test_obj : unittest subclass instance
    The unittest subclass instance to send the exceptions to.
  a1 : tf.tensor
    The tensor to be evaled then tested against the second array
  a2 : np.array
    Second array
  threshold : float
    The allowed difference in values in a1 and a2 to still be considered equal

  """
  # Initialize any variables that may exist in the graph
  init_op = tf.group(
    tf.global_variables_initializer(),
    tf.local_variables_initializer()
  )

  # Run the session to eval the tensor.
  with tf.Session() as sess:
    sess.run(init_op)
    a = sess.run(a1)
  if to_list:
    a = a.tolist()
  print a


def create_test_skeleton(file_name, class_name, module_key):
  """Creates a skeleton of the unittest of some particular module with a lot of the standard information already filled in.

  Parameters
  ----------
  file_name : str
    The file name of the module to be unittested
  class_name : str
    Name of the unittest class (usually just Test<Camel case version of file_name>)
  module_key : the alias of the module (usually just a two letter shortening of the file_name)
    Description of parameter `module_key`.

  Returns
  -------
  type
    Description of returned object.

  """

  # Define the template unittest file
  template = """import <OBJ_MOD> as o
import shutil
import tempfile
import unittest
import production.utils.test_helpers as th
import <MOD_PATH> as <KEY>
import os
import production.utils.dir_functions as d

this_path = os.path.dirname(os.path.realpath(__file__))

class <CLASS_NAME>(unittest.TestCase):
  def setUp(self):
    pass

  def tearDown(self):
    pass

  <TEST_FUNCTIONS>

if __name__ == "__main__":
  unittest.main()"""

  obj_mod_file_name = file_name.replace('.py', '_objects.py')
  obj_mod = obj_mod_file_name.replace('.py', '')

  unit_test_dir = d.maybe_create_dir('unit_test')
  with open(os.path.join(unit_test_dir, obj_mod_file_name), 'w') as obj_mod_file:
    obj_mod_file.write('')

  mod_path = os.getcwd()
  for pythonpath in os.environ['PYTHONPATH'].split(':'):
    mod_path = mod_path.replace(pythonpath, '')

  mod_path = mod_path.replace('/', '.').replace('.py', '').lstrip('.')
  mod_path += '.' + file_name.replace('.py', '')
  loaded = imp.load_source('*', os.path.join(os.getcwd(), file_name))

  func_tuples = []
  func_strs = []
  func_tuples.extend(i.getmembers(loaded, predicate=i.isfunction))

  for cls in i.getmembers(loaded, predicate=i.isclass):
    func_tuples.extend(i.getmembers(cls, predicate=i.ismethod))

  for func_name, func in func_tuples:
    n_args = func.__code__.co_argcount
    func_str = 'def test_' + func_name + '(self):'
    func_str += '\n    outputs = ' + module_key + '.' + func_name + '('
    func_str += '\n      ' + ',\n      '.join(func.__code__.co_varnames[:n_args])
    func_str += '\n    )'
    func_strs.append(func_str)

  test_func_str = '\n\n  '.join(func_strs)

  template = template.replace('<OBJ_MOD>', obj_mod)
  template = template.replace('<MOD_PATH>', mod_path)
  template = template.replace('<KEY>', module_key)
  template = template.replace('<CLASS_NAME>', class_name)
  template = template.replace('<TEST_FUNCTIONS>', test_func_str)

  test_file_name = 'test_' + file_name
  with open(os.path.join(unit_test_dir, test_file_name), 'w') as test_file:
    test_file.write(template)



"""GetItem tank definition."""
import wtrwrks.waterworks.waterwork_part as wp
import wtrwrks.waterworks.tank as ta
import wtrwrks.tanks.utils as ut


class GetItem(ta.Tank):
  """Run __getitem__ on some object (e.g. a dictionary) to return some value.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """

  func_name = 'getitem'
  slot_keys = ['a', 'key']
  tube_keys = ['target', 'a', 'key']
  pass_through_keys = ['a', 'key']

  def _pour(self, a, key):
    """

    Parameters
    ----------
    a: object
      The object to getitem from.
    key: hashable
      The key to pass to the getitem

    Returns
    -------
    dict(
      target: object
        The value returned from the __getitem__ call to 'a'.
      a: object
        The object to getitem from.
      key: hashable
        The key to pass to the getitem
    )

    """
    return {'target': a[key], 'a': ut.maybe_copy(a)}

  def _pump(self, target, a, key):
    """Execute the Shape tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    target: object
      The value returned from the __getitem__ call to 'a'.
    a: object
      The object to getitem from.
    key: hashable
      The key to pass to the getitem

    Returns
    -------
    dict(
      a: object
        The object to getitem from.
      key: hashable
        The key to pass to the getitem
    )

    """
    return {'a': ut.maybe_copy(a), 'key': key}

  



"""Mul tank definition."""
import wtrwrks.waterworks.tank as ta
import wtrwrks.tanks.utils as ut
import numpy as np


class Mul(ta.Tank):
  """The defintion of the Mul tank. Contains the implementations of the _pour and _pump methods, as well as which slots and tubes the waterwork objects will look for.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """

  func_name = 'mul'
  slot_keys = ['a', 'b']
  tube_keys = ['target', 'smaller_size_array', 'a_is_smaller', 'missing_vals']

  def _pour(self, a, b):
    """Execute the Mul tank (operation) in the pour (forward) direction.

    Parameters
    ----------
    a: np.ndarray
      The first array to be multiplied
    b: np.ndarray
      The second array to be multiplied

    Returns
    -------
    dict(
      target: np.ndarray
        The result of a*b
      smaller_size_array: np.ndarray
        Either 'a' or 'b' depending on which has fewer elements.
      a_is_smaller: bool
        Whether a is the smaller sized array.
      missing_vals: np.ndarray
        The values from either 'a' or 'b' that were lost when the other array had a zero in that location.
    )

    """
    # If a or b is not a numpy array, then cast them to it.
    if type(a) is not np.ndarray:
      a = np.array(a)
    if type(b) is not np.ndarray:
      b = np.array(b)

    # Save the array which has a fewer number of elements. Since we can
    # reconstruct the original shape of the larger array from the target.
    a_is_smaller = a.size < b.size
    if a_is_smaller:
      smaller_size_array = ut.maybe_copy(a)
    else:
      smaller_size_array = ut.maybe_copy(b)

    # Multiply them together and save all the values which were effectively
    # 'erased' by a corresponding zero in the smaller array. We don't need to
    # to it for the other array since the smaller sized array is going to be
    # saved anyway.
    target = np.array(a * b)
    if a_is_smaller:
      missing_vals = b[target == 0]
    else:
      missing_vals = a[target == 0]

    return {'target': target, 'smaller_size_array': smaller_size_array, 'a_is_smaller': a_is_smaller, 'missing_vals': missing_vals}

  def _pump(self, target, smaller_size_array, a_is_smaller, missing_vals):
    """Execute the Mul tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    target: np.ndarray
      The result of a*b
    smaller_size_array: np.ndarray
      Either 'a' or 'b' depending on which has fewer elements.
    a_is_smaller: bool
      Whether a is the smaller sized array.
    missing_vals: np.ndarray
      The values from either 'a' or 'b' that were lost when the other array had a zero in that location.

    Returns
    -------
    dict(
      a: np.ndarray
        The first array to be multiplied
      b: np.ndarray
        The second array to be multiplied
    )

    """
    # Find the value of the larger array using target and the smaller array.
    # Fill in any missing values which occured when there was a zero involved.
    if a_is_smaller:
      a = ut.maybe_copy(smaller_size_array)
      b = np.array(target / a)
      b[target == 0] = missing_vals
    else:
      a = np.array(target / smaller_size_array)
      b = ut.maybe_copy(smaller_size_array)
      a[target == 0] = missing_vals
    return {'a': a, 'b': b}



import unittest
import wtrwrks.utils.test_helpers as th
import wtrwrks.tanks.tank_defs as td
import numpy as np


class TestReplace(th.TestTank):

  def test_one_d(self):
    self.pour_pump(
      td.replace,
      {
        'a': np.array([1, 3]),
        'mask': np.array([True, False]),
        'replace_with': np.array([0])
      },
      {
        'target': np.array([0, 3]),
        'mask': np.array([True, False]),
        'replaced_vals': np.array([1, 0]),
        'replace_with': np.array([0])
      },
      type_dict={'a': np.ndarray, 'mask': np.ndarray, 'replace_with': np.ndarray},
      test_type=False
    )

  def test_two_d(self):
    self.pour_pump(
      td.replace,
      {
        'a': np.array([[0, 1], [2, 3], [4, 5], [1, 0]]),
        'mask': np.array([[0, 1], [1, 1], [0, 1], [1, 0]]).astype(bool),
        'replace_with': np.array([6, 7, 6, 7, 6])
      },
      {
        'target': np.array([[0, 6], [7, 6], [4, 7], [6, 0]]),
        'mask': np.array([[0, 1], [1, 1], [0, 1], [1, 0]]).astype(bool),
        'replaced_vals': np.array([[0, 1], [2, 3], [0, 5], [1, 0]]),
        'replace_with': np.array([6, 7, 6, 7, 6])
      },
      type_dict={'a': np.ndarray, 'mask': np.ndarray, 'replace_with': np.ndarray}
    )

    self.pour_pump(
      td.replace,
      {
        'a': np.array([[0, 1], [2, 3], [4, 5]]),
        'mask': np.array([True, False, False]),
        'replace_with': np.array([-1])
      },
      {
        'target': np.array([[-1, -1], [2, 3], [4, 5]]),
        'mask': np.array([True, False, False]),
        'replaced_vals': np.array([[0, 1], [0, 0], [0, 0]]),
        'replace_with': np.array([-1])
      },
      type_dict={'a': np.ndarray, 'mask': np.ndarray, 'replace_with': np.ndarray},
      test_type=False
    )

  def test_three_d(self):
    target = np.concatenate([np.arange(18).reshape((3, 3, 2)), np.zeros((1, 3, 2))], axis=0)
    self.pour_pump(
      td.replace,
      {
        'a': np.arange(24, dtype=float).reshape((4, 3, 2)),
        'mask': np.array([False, False, False, True]),
        'replace_with': np.array([0])
      },
      {
        'target': target,
        'mask': np.array([False, False, False, True]),
        'replaced_vals': np.array([[[0.0, 0.0], [0.0, 0.0], [0.0, 0.0]], [[0.0, 0.0], [0.0, 0.0], [0.0, 0.0]], [[0.0, 0.0], [0.0, 0.0], [0.0, 0.0]], [[18.0, 19.0], [20.0, 21.0], [22.0, 23.0]]], dtype=float),
        'replace_with': np.array([0])
      },
      type_dict={'a': np.ndarray, 'mask': np.ndarray, 'replace_with': np.ndarray},
      test_type=False
    )

if __name__ == "__main__":
    unittest.main()



import wtrwrks.waterworks.globs as gl
import wtrwrks.waterworks.waterwork_part as wp
import wtrwrks.waterworks.name_space as ns
import wtrwrks.utils.dir_functions as d
import wtrwrks.utils.multi as mu
import wtrwrks.utils.batch_functions as b
from wtrwrks.waterworks.empty import Empty, empty
import wtrwrks.read_write.tf_features as feat
import os
import pprint
import importlib
import dill as pickle
import tensorflow as tf
import numpy as np
import logging
import pathos.multiprocessing as mp
import glob
import re
import itertools
import traceback
import jpype


class Waterwork(object):
  """The full graph of tanks (i.e. operations) on the data, along with all slots and tubes which define the inputs/outputs of operations and hold their values. Can be thought of as a larger reversible operation that are composed of many smaller reversible operations.

  Attributes
  ----------
  funnels : dict(
    keys - strs. Names of the funnels.
    values - Slot objects.
  )
    All of the slots defined within the waterwork which are not connected to some other tube. i.e. the 'open' slots that need data in order to produce an output in the pour direction.
  taps : dict(
    keys - strs. Names of the taps.
    values - Tube objects.
  )
    All of the tubes defined within the waterwork which are not connected to some other slot. i.e. the 'open' tubes that need data in order to produce an output in the pump direction.
  slots : dict(
    keys - strs. Names of the slots.
    values - Slot objects.
  )
    All of the slots defined within the waterwork.
  tubes : dict(
    keys - strs. Names of the tubes.
    values - Tube objects.
  )
    All of the tubes defined within the waterwork.
  tanks : dict(
    keys - strs. Names of the tanks.
    values - Tube objects.
  )
    All of the tanks (or operations) defined within the waterwork.
  """

  def __init__(self, name='', from_file=None):
    """Initialize the waterwork to have empty funnels, slots, tanks, and taps."""
    self.funnels = {}
    self.tubes = {}
    self.slots = {}
    self.tanks = {}
    self.taps = {}
    self.merged = {}
    self.name = name

    if from_file is not None:
      save_dict = d.read_from_file(from_file)
      self._from_save_dict(save_dict)

  def __enter__(self):
    """When entering, set the global _default_waterwork to this waterwork."""
    if gl._default_waterwork is not None:
      raise ValueError("_default_waterwork is already set. Cannot be reset until context is exitted. Are you within the with statement of another waterwork?")

    # Create a new namespace for this waterwork
    self.name_space = ns.NameSpace(self.name)
    self.name_space.__enter__()

    gl._default_waterwork = self
    return self

  def __exit__(self, exc_type, exc_val, exc_tb):
    """When exiting, set the global _default_waterwork back to None."""
    gl._default_waterwork = None
    self.name_space.__exit__(exc_type, exc_val, exc_tb)

  def _from_save_dict(self, save_dict):
    import wtrwrks.tanks.tank_defs as td
    with ns.NameSpace(''):
      self.name = save_dict['name']
      for tank_name in save_dict['tanks']:
        tank_dict = save_dict['tanks'][tank_name]
        func = getattr(td, tank_dict['func_name'])

        kwargs = {}
        if 'kwargs' in tank_dict:
          kwargs = tank_dict['kwargs']

        args = []
        if 'args' in tank_dict:
          args = tank_dict['args']

        tubes, slots = func(name=tank_name, waterwork=self, *args, **kwargs)
        tank = tubes[tubes.keys()[0]].tank
        self.tanks[tank_name] = tank

      for slot_name in save_dict['slots']:
        slot_dict = save_dict['slots'][slot_name]
        tank = self.tanks[slot_dict['tank']]

        slot = tank.get_slot(slot_dict['key'])
        slot.plug = slot_dict['plug']

        # Set to proper name
        del self.slots[slot.name]
        slot.name = slot_name
        self.slots[slot_name] = slot

      for tube_name in save_dict['tubes']:
        tube_dict = save_dict['tubes'][tube_name]
        tank = self.tanks[tube_dict['tank']]

        tube = tank.get_tube(tube_dict['key'])
        tube.plug = tube_dict['plug']

        # Set to proper name
        del self.tubes[tube.name]
        tube.name = tube_name
        self.tubes[tube_name] = tube

      for tube_name in save_dict['tubes']:
        tube = self.tubes[tube_name]
        downstream_tube_name = save_dict['tubes'][tube_name]['downstream_tube']

        if downstream_tube_name is not None:
          tube.downstream_tube = self.tubes[downstream_tube_name]

      for slot_name in self.slots:
        slot_dict = save_dict['slots'][slot_name]
        slot = self.slots[slot_name]

        if slot_dict['tube'] is not None:
          tube = self.tubes[slot_dict['tube']]
          tube.slot = slot
        else:
          tube = empty

        slot.tube = tube

      self.funnels = {}
      for funnel_name in save_dict['funnels']:
        self.funnels[funnel_name] = self.slots[funnel_name]

      self.taps = {}
      for tap_name in save_dict['taps']:
        self.taps[tap_name] = self.tubes[tap_name]

    for key in save_dict['merged']:
      tube = self.tubes[key]
      self.merged[tube] = set([self.tubes[k] for k in save_dict['merged'][key]])

  def _pour_tank_order(self):
    """Get the order to calculate the tanks in the pour direction.

    Returns
    -------
    list of tank objects
        The tanks ordered in such a way that they are guaranteed to have all the information to perform the operation.

    """
    tanks = sorted([self.tanks[t] for t in self.tanks if not self.tanks[t].get_slot_tanks()])
    visited = set(tanks)
    sorted_tanks = []

    while tanks:
      tank = tanks.pop(0)
      sorted_tanks.append(tank)

      child_tanks = tank.get_tube_tanks() - visited
      for child_tank in child_tanks:
        # If there are any tanks child_tank depnends on , that haven't been
        # visited, then continue.

        if child_tank.get_slot_tanks() - visited:
          continue

        tanks.append(child_tank)
        visited.add(child_tank)

    return sorted_tanks

  def _pump_tank_order(self):
    """Get the order to calculate the tanks in the pump direction.

    Returns
    -------
    list of tank objects
        The tanks ordered in such a way that they are guaranteed to have all the information to perform the operation.

    """
    tanks = sorted([self.tanks[t] for t in self.tanks if not self.tanks[t].get_tube_tanks()])
    visited = set(tanks)
    sorted_tanks = []

    while tanks:
      tank = tanks.pop(0)
      sorted_tanks.append(tank)

      child_tanks = tank.get_slot_tanks() - visited
      for child_tank in child_tanks:
        # If there are any tanks child_tank depnends on , that haven't been
        # visited, then continue.
        if child_tank.get_tube_tanks() - visited:
          continue

        tanks.append(child_tank)
        visited.add(child_tank)
    # tanks = sorted([self.tanks[t] for t in self.tanks])
    # return sorted(tanks, cmp=lambda a, b: 0 if b in a.get_pump_dependencies() else -1)
    return sorted_tanks

  def _save_dict(self):
    save_dict = {}
    save_dict['name'] = self.name
    save_dict['funnels'] = sorted(self.funnels.keys())
    save_dict['taps'] = sorted(self.taps.keys())

    save_dict['tanks'] = {}
    save_dict['slots'] = {}
    save_dict['tubes'] = {}
    save_dict['merged'] = {}

    for key in self.tanks:
      tank = self.tanks[key]
      save_dict['tanks'][key] = tank._save_dict()

    for key in self.slots:
      slot = self.slots[key]
      save_dict['slots'][key] = slot._save_dict()

    for key in self.tubes:
      tube = self.tubes[key]
      save_dict['tubes'][key] = tube._save_dict()

    for tube in self.merged:
      key = tube.name
      save_dict['merged'][key] = sorted([t.name for t in self.merged[tube]])

    return save_dict

  def clear_vals(self):
    """Set all the slots, tubes and placeholder values back to None """
    for d in [self.slots, self.tubes]:
      for key in d:
        d[key].set_val(None)

  def maybe_get_slot(self, *args):
    """Get a particular tank's if it exists, otherwise return None. Can take a variety input types.

    Parameters
    ----------
    *args: Slot, str or tuple
      Either the slot itself, the name of the slot or the (tank name, slot key) tuple.

    Returns
    -------
    Slot or None
      The slot object if found, otherwise a None

    """
    import wtrwrks.waterworks.slot as sl
    # Pull out the tank and key depending on the type and number of inputs.
    if len(args) == 2:
      tank = args[0]
      key = args[1]
    elif len(args) == 1 and type(args[0]) is tuple:
      tank = args[0][0]
      key = args[0][1]
    elif len(args) == 1 and type(args[0]) in (str, unicode) and args[0] in self.slots:
      return self.slots[args[0]]
    elif len(args) == 1 and isinstance(args[0], sl.Slot):
      return args[0]
    else:
      return None

    # Pull out the relevant tank object.
    if type(tank) in (str, unicode) and tank in self.tanks:
      tank = self.tanks[tank]
    elif isinstance(tank, sl.Slot):
      pass
    else:
      return None

    # Get the slot
    if key in tank.slots:
      return tank.slots[key]

    return None

  def maybe_get_tube(self, *args):
    """Get a particular tank's if it exists, otherwise return None. Can take a variety input types.

    Parameters
    ----------
    *args: tube, str or tuple
      Either the tube itself, the name of the tube or the (tank name, tube key) tuple.

    Returns
    -------
    tube or None
      The tube object if found, otherwise a None

    """
    import wtrwrks.waterworks.tube as tu
    # Pull out the tank and key depending on the type and number of inputs.
    if len(args) == 2:
      tank = args[0]
      key = args[1]
    elif len(args) == 1 and type(args[0]) is tuple:
      tank = args[0][0]
      key = args[0][1]
    elif len(args) == 1 and type(args[0]) in (str, unicode) and args[0] in self.tubes:
      return self.tubes[args[0]]
    elif len(args) == 1 and isinstance(args[0], tu.Tube):
      return args[0]
    else:
      return None

    # Pull out the relevant tank object.
    if type(tank) in (str, unicode) and tank in self.tanks:
      tank = self.tanks[tank]
    elif isinstance(tank, tu.Tube):
      pass
    else:
      return None

    # Get the tube
    if key in tank.tubes:
      return tank.tubes[key]

    return None

  def get_slot(self, tank, key):
    """Get a particular tank's slot.

    Parameters
    ----------
    tank : Tank or str
        Either the tank object or the name of the tank.
    key : str
        The slot key of the slot for that tank.

    Returns
    -------
    Slot
        The slot object

    """
    if type(tank) in (str, unicode):
      tank = self.tanks[tank]

    return self.tanks[tank.name].slots[key]

  def get_tube(self, tank, key):
    """Get a particular tank's tube.

    Parameters
    ----------
    tank : Tank or str
        Either the tank object or the name of the tank.
    key : str
        The tube key of the tube for that tank.

    Returns
    -------
    Tube
        The tube object.

    """
    if type(tank) in (str, unicode):
      tank = self.tanks[tank]

    return self.tanks[tank.name].tubes[key]

  def get_taps(self, return_plugged=False):
    """Get a dictionary of all the taps. Optionally choose to return the taps which are plugged or not."""
    r_dict = {k: v for k, v in self.taps.iteritems() if v.plug is None}
    return r_dict

  def get_funnels(self, return_plugged=False):
    """Get a dictionary of all the funnels. Optionally choose to return the funnels which are plugged or not."""
    r_dict = {k: v for k, v in self.funnels.iteritems() if v.plug is None}
    return r_dict

  def merge_tubes(self, target, *args):
    self.merged[target] = set()
    for arg in args:
      if type(arg) is Empty:
        continue
      if arg in self.merged:
        for other_arg in self.merged[arg]:
          other_arg.downstream_tube = target
          self.merged[target].add(other_arg)
        del self.merged[arg]

      elif arg.downstream_tube is not None and arg.downstream_tube != target:
        self.merged[target].add(arg.downstream_tube)
        arg.downstream_tube.downstream_tube = target

        if arg.downstream_tube in self.merged:
          for equal_arg in self.merged[arg.downstream_tube]:
            equal_arg.downstream_tube = target
            self.merged[target].add(equal_arg)

          if arg.downstream_tube != target:
            del self.merged[arg.downstream_tube]

      self.merged[target].add(arg)
      arg.downstream_tube = target

  def multi_pour(self, funnel_dict_iter, key_type='tube', return_plugged=False, num_threads=1, use_threading=False, batch_size=None, pour_func=None):
    if pour_func is None:
      save_dict = self._save_dict()

      def pour_func(funnel_dict):
        ww = Waterwork()
        ww._from_save_dict(save_dict)

        tap_dict = ww.pour(funnel_dict, key_type, return_plugged)
        return tap_dict

    tap_dicts = []
    for batch_num, batch in enumerate(b.batcher(funnel_dict_iter, batch_size)):
      tap_dicts.extend(
        mu.multi_map(pour_func, batch, num_threads, use_threading)
      )

    return tap_dicts

  def multi_pump(self, tap_dict_iter, key_type='slot', return_plugged=False, num_threads=1, use_threading=False, batch_size=None, pump_func=None):
    if pump_func is None:
      save_dict = self._save_dict()

      def pump_func(tap_dict):
        ww = Waterwork()
        ww._from_save_dict(save_dict)

        funnel_dict = ww.pump(tap_dict, key_type, return_plugged)
        return funnel_dict

    funnel_dicts = []
    for batch_num, batch in enumerate(b.batcher(tap_dict_iter, batch_size)):
      funnel_dicts.extend(
        mu.multi_map(pump_func, batch, num_threads, use_threading)
      )

    return funnel_dicts

  def multi_write_examples(self, funnel_dict_iter, file_name, num_threads=1, use_threading=False, batch_size=None, file_num_offset=0, skip_fails=False, skip_keys=None, serialize_func=None):
    if serialize_func is None:
      save_dict = self._save_dict()

      def serialize_func(funnel_dict):
        jpype.attachThreadToJVM()
        ww = Waterwork()
        ww._from_save_dict(save_dict)

        tap_dict = ww.pour(funnel_dict, 'str', False)
        feature_dict, func_dict = self._get_feature_dicts(tap_dict)

        serial = ww._serialize_tap_dict(tap_dict, func_dict)
        return serial

    if type(funnel_dict_iter) in (list, tuple):
      funnel_dict_iter = (i for i in funnel_dict_iter)

    file_names = []
    if not file_name.endswith('.tfrecord'):
      raise ValueError("file_name must end in '.tfrecord'")

    dir = file_name.split('/')[:-1]
    d.maybe_create_dir(*dir)

    for batch_num, batch in enumerate(b.batcher(funnel_dict_iter, batch_size)):
      if batch_num == 0:
        tap_dict = self.pour(batch[0], key_type='str', return_plugged=False)
        feature_dict, func_dict = self._get_feature_dicts(tap_dict)
        feature_dict_fn = re.sub(r'_?[0-9]*.tfrecord', '.pickle', file_name)
        d.save_to_file(feature_dict, feature_dict_fn)

      logging.info("Serializing batch %s", batch_num)
      if skip_fails:
        try:
          all_serials = mu.multi_map(serialize_func, batch, num_threads, use_threading)
        except Exception:
          logging.warn("Batched %s failed. Skipping.", batch_num)
          continue
      else:
        all_serials = mu.multi_map(serialize_func, batch, num_threads, use_threading)
      logging.info("Finished serializing batch %s", batch_num)

      file_num = file_num_offset + batch_num
      fn = file_name.replace('.tfrecord', '_' + str(file_num) + '.tfrecord')
      file_names.append(fn)

      logging.info("Writing batch %s", batch_num)
      writer = tf.python_io.TFRecordWriter(fn)
      for serials in all_serials:
        for serial in serials:
          writer.write(serial)
      logging.info("Finished writing batch %s", batch_num)
      writer.close()

    return file_names

  def multi_read_examples(self, file_name_iter, num_threads=1, key_type='slot', return_plugged=False, use_threading=False, skip_fails=False):
    """Pours the arrays then writes the examples to tfrecords in a multithreading manner. It creates one example per 'row', i.e. axis=0 of the arrays. All arrays must have the same axis=0 dimension and must be of a type that can be written to a tfrecord

    Parameters
    ----------
    funnel_dicts : list of dicts(
      keys - Slot objects or Placeholder objects. The 'funnels' (i.e. unconnected slots) of the waterwork.
      values - valid input data types
    )
      The inputs to the waterwork's full pour functions. There is exactly one funnel_dict for every process.
    file_name : str
      The name of the tfrecord file to write to. An extra '_<num>' will be added to the name.
    file_num_offset : int
      A number that controls what number will be appended to the file name (so that files aren't overwritten.)

    """
    # tap_dicts = []
    file_names = []
    for file_name in file_name_iter:
      file_names.append(file_name)
    tap_dicts = self._files_to_tap_dicts(file_names)

    funnel_dicts = []
    if skip_fails:
      try:
        funnel_dicts = self.multi_pump(tap_dicts, num_threads, key_type, return_plugged, use_threading)
      except Exception:
        logging.warn("Batched %s failed. Skipping.", file_names)
        funnel_dicts = []
    else:
      funnel_dicts = self.multi_pump(tap_dicts, num_threads, key_type, return_plugged, use_threading)

    return funnel_dicts

  def pour(self, funnel_dict=None, key_type='tube', return_plugged=False):
    """Run all the operations of the waterwork in the pour (or forward) direction.

    Parameters
    ----------
    funnel_dict : dict(
      keys - Slot objects or Placeholder objects. The 'funnels' (i.e. unconnected slots) of the waterwork.
      values - valid input data types
    )
        The inputs to the waterwork's full pour function.
    key_type : str ('tube', 'tuple', 'name')
      The type of keys to return in the return dictionary. Can either be the tube objects themselves (tube), the tank, output key pair (tuple) or the name (str) of the tube.

    Returns
    -------
    dict(
      keys - Tube objects, (or tuples if tuple_keys set to True). The 'taps' (i.e. unconnected tubes) of the waterwork.
    )
        The outputs of the waterwork's full pour function

    """

    if funnel_dict is None:
      funnel_dict = {}

    # Set all the values of the funnels from the inputted arguments.
    evaled_dict = {}
    logging.debug("%s's funnel_dict - %s", self.name, sorted(funnel_dict))
    for ph, val in funnel_dict.iteritems():
      sl_obj = self.maybe_get_slot(ph)
      if sl_obj is not None:
        if sl_obj.plug is not None:
          raise ValueError(str(sl_obj) + ' has a plug. If you want to set the value dynamically then do funnel.unplug().')

        evaled_dict[sl_obj.name] = val
        sl_obj.set_val(val)
        if type(sl_obj.tube) is not Empty:
          sl_obj.tube.set_val(val)
      else:
        raise ValueError(str(ph) + ' is not a supported input into pour function')

    # Check that all funnels have a value
    logging.debug("%s's funnels - %s", self.name, sorted(self.funnels))
    for funnel_key in self.funnels:
      logging.debug("funnel_key - %s", funnel_key)
      funnel = self.funnels[funnel_key]
      if funnel.plug is not None:
        logging.debug("Plugging - %s", funnel_key)
        # funnel.set_val(funnel.plug(stand_funnel_dict))
      elif funnel.get_val() is None:
        raise ValueError("All funnels must have a set value. " + str(funnel) + " is not set.")

    # Run all the tanks (operations) in the pour direction, filling all slots'
    # and tubes' val attributes as you go.
    tanks = self._pour_tank_order()
    logging.debug("%s's pour_tank_order - %s", self.name, [t.name for t in tanks])
    for tank in tanks:
      logging.debug("Pouring tank - %s", tank.name)

      # Fill any of the slots that are to be plugged
      kwargs = {}
      for slot_name in tank.slots:
        slot = tank.slots[slot_name]
        if slot.plug is not None:
          slot.set_val(slot.plug(evaled_dict))

      kwargs = {k: tank.slots[k].get_val() for k in tank.slots}

      logging.debug("Inputting kwargs to pour - %s", {k: v for k, v in kwargs.iteritems()})

      try:
        tube_dict = tank.pour(**kwargs)
      except:
        logging.exception("Failure in pour of tank %s", tank.name)
        raise

      for key in tube_dict:
        slot = tank.tubes[key].slot

        if type(slot) is not Empty:
          slot.set_val(tube_dict[key])
          evaled_dict[slot.name] = tube_dict[key]
    # Create the dictionary to return
    r_dict = {}
    logging.debug("%s's taps - %s", self.name, sorted(self.taps))
    for tap_name in self.taps:
      logging.debug("setting tap_key - %s", tap_name)
      tap = self.taps[tap_name]

      if tap.plug is not None and not return_plugged:
        continue

      if key_type == 'tube':
        r_dict[tap] = tap.get_val()
      elif key_type == 'tuple':
        r_dict[tap.get_tuple()] = tap.get_val()
      elif key_type == 'str':
        r_dict[tap.name] = tap.get_val()
      else:
        raise ValueError(str(key_type) + " is an invalid key_type.")

    return r_dict

  def pump(self, tap_dict=None, key_type='slot', return_plugged=False):
    """Run all the operations of the waterwork in the pump (or backward) direction.

    Parameters
    ----------
    tap_dict : dict(
      keys - Tube objects. The 'taps' (i.e. unconnected tubes) of the waterwork.
    )
        The inputs of the waterwork's full pump function
    key_type : str ('tube', 'tuple', 'name')
      The type of keys to return in the return dictionary. Can either be the tube objects themselves (tube), the tank, output key pair (tuple) or the name (str) of the tube.

    Returns
    -------
    dict(
      keys - Slot objects. The 'funnels' (i.e. unconnected slots) of the waterwork.
      values - valid input data types
    )
        The outputs to the waterwork's full pump function.

    """
    if tap_dict is None:
      tap_dict = {}

    evaled_dict = {}
    # Set all the values of the taps from the inputted arguments.
    logging.debug("%s's tap_dict - %s", self.name, sorted(tap_dict))
    for tap, val in tap_dict.iteritems():
      tu_obj = self.maybe_get_tube(tap)
      if tu_obj is not None:
        if tu_obj.downstream_tube is not None:
          logging.warn("%s has downstream_tube %s. Setting that value instead.", tu_obj.name, tu_obj.downstream_tube.name)
          tu_obj = tu_obj.downstream_tube
        if tu_obj.plug is not None:
          raise ValueError(str(tu_obj) + ' has a plug. Cannot set the value of a tap which is plugged.')

        evaled_dict[tu_obj.name] = val
        tu_obj.set_val(val)
      else:
        raise ValueError(str(tap) + ' is not a supported form of input into pump function')

    # Check that all funnels have a value
    logging.debug("%s's taps - %s", self.name, sorted(self.taps))
    for tap_key in self.taps:
      logging.debug("tap_key - %s", tap_key)
      tap = self.taps[tap_key]
      if tap.plug is not None:
        logging.debug("Plugging - %s", tap_key)
        tap.set_val(tap.plug(evaled_dict))
      elif tap.get_val() is None:
        raise ValueError("All taps must have a set value. " + str(tap) + " is not set.")

    # Run all the tanks (operations) in the pump direction, filling all slots'
    # and tubes' val attributes as you go.
    tanks = self._pump_tank_order()
    logging.debug("%s's pump_tank_order - %s", self.name, [t.name for t in tanks])
    for tank in tanks:
      logging.debug("Pumping tank - %s", tank.name)
      for tube_name in tank.tubes:
        tube = tank.tubes[tube_name]
        if tube.plug is not None:
          tube.set_val(tube.plug(evaled_dict))
      kwargs = {k: tank.tubes[k].get_val() for k in tank.tubes}

      logging.debug("Inputting kwargs to pour - %s", {k: v for k, v in kwargs.iteritems()})

      try:
        slot_dict = tank.pump(**kwargs)
      except:
        logging.exception("Failure in pump of tank %s", tank.name)
        raise

      for key in slot_dict:
        tube = tank.slots[key].tube

        if type(tube) is not Empty:
          tube.set_val(slot_dict[key])
          evaled_dict[tube.name] = slot_dict[key]
    # Create the dictionary to return
    r_dict = {}
    logging.debug("%s's funnels - %s", self.name, sorted(self.funnels))
    for funnel_name in self.funnels:
      funnel = self.funnels[funnel_name]

      if funnel.plug is not None and not return_plugged:
        continue

      if key_type == 'slot':
        r_dict[funnel] = funnel.get_val()
      elif key_type == 'tuple':
        r_dict[funnel.get_tuple()] = funnel.get_val()
      elif key_type == 'str':
        r_dict[funnel.name] = funnel.get_val()
      else:
        raise ValueError(str(key_type) + " is an invalid key_type.")

    return r_dict

  def read_and_decode(self, serialized_example, feature_dict, prefix=''):
    """Convert a serialized example created from an example dictionary from this transform into a dictionary of shaped tensors for a tensorflow pipeline.

    Parameters
    ----------
    serialized_example : tfrecord serialized example
      The serialized example to read and convert to a dictionary of tensors.
    feature_dict :
      A dictionary with the keys being the keys of the tap dict (and their shapes) and the values being the FixedLenFeature's with the proper values all filled in

    Returns
    -------
    dict of tensors
      The tensors created by decoding the serialized example and reshaping them.

    """
    feat_dict = {}
    for key in feature_dict:
      shape = feature_dict[key]['shape']
      tf_dtype = feature_dict[key]['tf_dtype']
      feat_dict[key] = tf.FixedLenFeature(shape, tf_dtype)

    features = tf.parse_single_example(
      serialized_example,
      features=feat_dict
    )

    return features

  def read_examples(self, file_name, key_type='slot', return_plugged=False):
    tap_dicts = self._files_to_tap_dicts([file_name])
    return self.pump(tap_dicts[0], key_type=key_type, return_plugged=return_plugged)

  def _files_to_tap_dicts(self, file_names):
    """Run the pump transformation on a list of example dictionaries to reconstruct the original array.

    Parameters
    ----------
    batched_arrays: dict of arrays
      The keys

    Returns
    -------
    np.ndarray
      The numpy array to transform into examples.

    """
    if not file_names:
      return []

    feature_dict_fn = re.sub(r'_?[0-9]*.tfrecord', '.pickle', file_names[0])

    if not os.path.isfile(feature_dict_fn):
      raise ValueError("Expected a feature_dict file named", feature_dict_fn)

    feature_dict = d.read_from_file(feature_dict_fn)

    tap_dicts = []
    for file_name in file_names:
      dataset = tf.data.TFRecordDataset(file_name)
      dataset = dataset.map(lambda a: self.read_and_decode(a, feature_dict))

      iter = tf.data.Iterator.from_structure(
        dataset.output_types,
        dataset.output_shapes
      )

      init = iter.make_initializer(dataset)
      features = iter.get_next()

      with tf.Session() as sess:
        sess.run(init)

        tap_dict = {}
        try:
          while True:
            example_dict = sess.run(features)
            for key in example_dict:
              tap_dict.setdefault(key, [])
              tap_dict[key].append(example_dict[key])

        except tf.errors.OutOfRangeError:
          pass

      for key in tap_dict:
        tap_dict[key] = np.stack(tap_dict[key], axis=0)

      for key in sorted(feature_dict, reverse=True):
        np_dtype = feature_dict[key]['np_dtype']

        if np_dtype.char == 'U':
          tap_dict[key] = tap_dict[key].astype(str)
          tap_dict[key] = np.char.decode(tap_dict[key], encoding='utf-8')
        elif np_dtype.char == 'S':
          tap_dict[key] = tap_dict[key].astype(np.str)
        else:
          tap_dict[key] = tap_dict[key].astype(np_dtype)
      tap_dicts.append(tap_dict)

    return tap_dicts

  def save_to_file(self, file_name):
    if not file_name.endswith('pickle') and not file_name.endswith('pkl') and not file_name.endswith('dill'):
      raise ValueError("Waterwork can only be saved as a pickle.")
    save_dict = self._save_dict()
    d.save_to_file(save_dict, file_name)

  def _write_tap_dict(self, writer, tap_dict, func_dict, skip_keys=None):
    if skip_keys is None:
      skip_keys = []

    keys_to_write = [k for k in tap_dict if k not in skip_keys]

    num_examples = None
    for key in keys_to_write:
      array = tap_dict[key]
      num_examples = array.shape[0]
      break

    for row_num in xrange(num_examples):
      example_dict = {}
      for key in keys_to_write:
        flat = tap_dict[key][row_num].flatten()
        example_dict[key] = func_dict[key](flat)

      example = tf.train.Example(
        features=tf.train.Features(feature=example_dict)
      )
      writer.write(example.SerializeToString())

  def _serialize_tap_dict(self, tap_dict, func_dict, skip_keys=None):
    if skip_keys is None:
      skip_keys = []

    keys_to_write = [k for k in tap_dict if k not in skip_keys]

    num_examples = None
    for key in keys_to_write:
      array = tap_dict[key]
      num_examples = array.shape[0]
      break

    serials = []
    for row_num in xrange(num_examples):
      example_dict = {}
      for key in keys_to_write:
        flat = tap_dict[key][row_num].flatten()
        example_dict[key] = func_dict[key](flat)

      example = tf.train.Example(
        features=tf.train.Features(feature=example_dict)
      )
      serials.append(example.SerializeToString())
    return serials

  def _get_feature_dicts(self, tap_dict):
    func_dict = {}
    feature_dict = {}
    for key in tap_dict:
      array = tap_dict[key]
      num_examples = array.shape[0]

      if array.shape[0] != num_examples:
        raise ValueError("All arrays must have the same size first dimesion in order to split them up into individual examples")

      if array.dtype not in (np.int32, np.int64, np.bool, np.float32, np.float64) and array.dtype.type not in (np.string_, np.unicode_):
        raise TypeError("Only string and number types are supported. Got " + str(array.dtype))
      feature_dict[key] = {}
      feature_dict[key]['np_dtype'] = array.dtype
      feature_dict[key]['tf_dtype'] = feat.select_tf_dtype(array.dtype)
      feature_dict[key]['shape'] = array.shape[1:]
      func_dict[key] = feat.select_feature_func(array.dtype)

    return feature_dict, func_dict

  def write_tap_dicts(self, tap_dicts, file_name, skip_keys=None):

    feature_dict, func_dict = self._get_feature_dicts(tap_dicts[0])

    writer = tf.python_io.TFRecordWriter(file_name)
    for tap_dict in tap_dicts:
      if not tap_dict:
        continue
      self._write_tap_dict(writer, tap_dict, func_dict, skip_keys)

    feature_dict_fn = re.sub(r'_?[0-9]*.tfrecord', '.pickle', file_name)
    d.save_to_file(feature_dict, feature_dict_fn)

    writer.close()

  def write_examples(self, funnel_dict, file_name):
    """Pours the array then writes the examples to tfrecords. It creates one example per 'row', i.e. axis=0 of the arrays. All arrays must have the same axis=0 dimension and must be of a type that can be written to a tfrecord

    Parameters
    ----------
    funnel_dict : dict(
      keys - Slot objects or Placeholder objects. The 'funnels' (i.e. unconnected slots) of the waterwork.
      values - valid input data types
    )
        The inputs to the waterwork's full pour function.
    file_name : str
      The name of the tfrecord file to write to.

    """
    if not file_name.endswith('.tfrecord'):
      raise ValueError("file_name must end in '.tfrecord'. Got: ", file_name)

    dir = file_name.split('/')[:-1]
    d.maybe_create_dir(*dir)

    writer = tf.python_io.TFRecordWriter(file_name)
    tap_dict = self.pour(funnel_dict, key_type='str')

    feature_dict, func_dict = self._get_feature_dicts(tap_dict)

    self._write_tap_dict(writer, tap_dict, func_dict)

    feature_dict_fn = re.sub(r'_?[0-9]*.tfrecord', '.pickle', file_name)
    d.save_to_file(feature_dict, feature_dict_fn)
    writer.close()

    return file_name


def recon_and_pump(args):
  # jpype.attachThreadToJVM()
  save_dict = args[0]
  tap_dict = args[1]
  key_type = args[2]
  return_plugged = args[3]

  ww = Waterwork()
  ww._from_save_dict(save_dict)

  funnel_dict = ww.pump(tap_dict, key_type, return_plugged)
  return funnel_dict


def inf_gen(val):
  while True:
    yield val



"""Lemmatize tank definition."""
import wtrwrks.waterworks.tank as ta
import wtrwrks.tanks.utils as ut
import wtrwrks.string_manipulations.diff as di
import numpy as np
# from chop.mmseg import Tokenizer as MMSEGTokenizer
# from chop.hmm import Tokenizer as HMMTokenizer


class Lemmatize(ta.Tank):
  """The defintion of the Lemmatize tank. Contains the implementations of the _pour and _pump methods, as well as which slots and tubes the waterwork objects will look for.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """

  func_name = 'lemmatize'
  slot_keys = ['strings', 'lemmatizer']
  tube_keys = ['target', 'lemmatizer', 'diff']
  pass_through_keys = ['lemmatizer']


  def _pour(self, strings, lemmatizer):
    """Execute the Lemmatize tank (operation) in the pour (forward) direction.

    Parameters
    ----------
    strings: np.ndarray of strings
      The array of strings to be lemmatized.
    lemmatizer: func
      A function which takes in a string and outputs a standardized version of that string

    Returns
    -------
    dict(
      target: np.ndarray of strings
        The array of lemmatized strings
      lemmatizer: func
        A function which takes in a string and outputs a standardized version of that string.
      diff: np.ndarray of strings
        The array of strings which define the differences between the original string and the string that has been lemmatized.
    )

    """
    strings = np.array(strings)
    target = np.vectorize(lemmatizer)(strings)

    diff = np.vectorize(di.get_diff_string)(target, strings)

    return {'target': target, 'diff': diff, 'lemmatizer': lemmatizer}

  def _pump(self, target, diff, lemmatizer):
    """Execute the Lemmatize tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    target: np.ndarray of strings
      The array of lemmatized strings
    lemmatizer: func
      A function which takes in a string and outputs a standardized version of that string.
    diff: np.ndarray of strings
      The array of strings which define the differences between the original string and the string that has been lemmatized.

    Returns
    -------
    dict(
      strings: np.ndarray of strings
        The array of strings to be lemmatized.
      lemmatizer: func
        A function which takes in a string and outputs a standardized version of that string
    )

    """
    strings = np.vectorize(di.reconstruct)(target, diff)

    return {'strings': strings, 'lemmatizer': lemmatizer}



"""Tank definition."""
import wtrwrks.waterworks.globs as gl
import wtrwrks.waterworks.waterwork_part as wp
from wtrwrks.waterworks.empty import Empty, empty
import wtrwrks.waterworks.slot as sl
import wtrwrks.waterworks.tube as tu
import os
import sys


class Tank(wp.WaterworkPart):
  """Base class for any tank, i.e. an operation that pulls in information from the 'slots', does some processing and outputs them to the 'tubes'.

  Attributes
  ----------
  waterwork : Waterwork or None
    The waterwork that the part will be added to. If None, it is assinged to the _default_waterwork.
  name : str or None
    The name of the part within the waterwork. Must be unique. If None, it will be set to a default value depending on the subclass.
  slot_keys : list of str
    The tank's (operation's) argument keys. They define the names of the inputs to the tank.
  tube_dict : dict(
    keys - strs. The tank's (operation's) output keys. THey define the names of the outputs of the tank
    values - types. The types of the arguments outputs.
  )
    The tank's (operation's) output keys and their corresponding types.
  slots : dict(
    keys - strs. equal to that of the slot_keys.
    values - Slot object.
  )
    The slot objects that define the pour direction inputs (or pump direction outputs) of the tank.
  tubes : dict(
    keys - strs. equal to that of the tube_keys.
    values - Slot object.
  )
    The tube objects that define the pour direction outputs (or pump direction inputs) of the tank.

  """
  func_name = None
  slot_keys = None
  tube_keys = None
  pass_through_keys = None

  def __init__(self, waterwork=None, name=None, **input_dict):
    """Create a Tank. Eagerly run the pour function if all the input values are known at creation.

    Parameters
    ----------
    waterwork : type
        Description of parameter `waterwork`.

    name : type
        Description of parameter `name`.

    **input_dict : kwargs = {
        keys - Slot keys. Must be the same as the attribute slot_keys.
        values - Tube, None or some valid input data type.
      }
      The inputs to the tank. If the value is of Tube type then the slot referenced by the key is 'connected' to that tube. If the value is None then it will be left empty and will become a 'funnel' (i.e. a slot that needs inputted data to run.) when running the Waterwork the tank belongs to. If all the values are either valid input data types or tubes with the values filled, then it will eagerly run the tank and fill all the relevant tubes.

    """
    # Use the WaterworkPart constructor to assign the waterwork and the name.
    super(Tank, self).__init__(waterwork, name)

    if self.name in self.waterwork.tanks:
      raise ValueError(self.name + " already defined as tank. Choose a different name.")

    # Assign self the the waterwork's directory of tanks.
    self.waterwork.tanks[self.name] = self

    # Make sure all the inputed arguments are valid.
    for key in input_dict:
      if key not in self.slot_keys:
        raise TypeError(key + ' not a valid argument for ' + str(type(self)))

    # Create all the slots and tubes for this tank.
    self.slots = {}
    self.tubes = {}
    self._create_slots(self.slot_keys, self.waterwork)
    self._create_tubes(self.tube_keys, self.waterwork)

    # Join the this tank's slots to the tubes of the other tanks which are
    # inputted in as an argument (input_dict).
    for key in input_dict:
      if type(input_dict[key]) is sl.Slot:
        raise ValueError("Cannot pass slot as argument to " + str(type(self)))
      if type(input_dict[key]) is list or type(input_dict[key]) is tuple:
        input_dict[key] = self._handle_iterable(input_dict[key])
      if type(input_dict[key]) is tu.Tube and input_dict[key].downstream_tube is not None:
        input_dict[key] = input_dict[key].downstream_tube

    self._join_tubes_to_slots(input_dict, self.waterwork)

    # If all the slots of the tank are 'filled', i.e. are either connected to a
    # tube with a non None val or are given a valid datum as input, then
    # eagerly run the tank's pour function and output the results to the tank's
    # tubes' vals.

    all_slots_filled = self._check_slots_filled(input_dict)

    for key in input_dict:
      if type(input_dict[key]) is sl.Slot:
        raise ValueError("Cannot pass slot as argument to " + str(type(self)))

      # If data is directly inputted into the slot, then create a placeholder
      # with it's value set to the data. And set it as plugged to that value by # default.
      if type(input_dict[key]) is not tu.Tube and type(input_dict[key]) is not Empty:
        self.slots[key].set_val(input_dict[key])
        self.slots[key].set_plug(input_dict[key], obj_is_callable=callable(input_dict[key]))
        self.slots[key].tube = empty

    if all_slots_filled:
      input_dict = self._convert_tubes_to_vals(input_dict)
      self.pour(**input_dict)

    import wtrwrks.tanks.clone as cl
    if isinstance(self, cl.MergeEqual):
      self._handle_merge()

    if self.pass_through_keys is not None:
      import wtrwrks.tanks.tank_defs as td
      for key in self.pass_through_keys:
        slot = self.slots[key]
        if type(slot.tube) is not Empty:
          td.merge_equal(slot.tube, self.tubes[key])
        if slot.plug is not None:
          self.tubes[key].plug = slot.plug

  def __hash__(self):
    """Uniquely identify the tank among other tanks in the waterwork."""
    return hash(self.name)

  def __getitem__(self, key):
    """Return a tube object of the tank, denonted by the key."""
    return self.tubes[key]

  def __str__(self):
    """Return the str of the tank, which is just its name."""
    return str(self.name)

  def _check_slots_filled(self, input_dict):
    """Check whether or not the all the values of the input_dict are set to determine whether or not the tank can be eagerly executed.

    Parameters
    ----------
    input_dict : dict(
        keys - Slot keys. Must be the same as the attribute slot_keys.
        values - Tube, None or some valid input data type.
      )
      The inputs to the tank.

    Returns
    -------
    bool
        Whether or not all the values needed for eager execution have been filled.

    """
    # If there are any slot keys that are not covered by the input_dict, then
    # they have not all been filled.
    if set(self.slot_keys) - set(input_dict.keys()):
      return False

    # Go through each value of the input dict. If there are any Nones, or there
    # are any Slot objects that don't have a value, then return False.
    # Otherwise return True.
    all_slots_filled = True
    for key in input_dict:
      if type(input_dict[key]) is Empty:
        all_slots_filled = False
        break
      if (
        type(input_dict[key]) is tu.Tube and
        input_dict[key].get_val() is None):
        all_slots_filled = False
        break
      if type(input_dict[key]) is list or type(input_dict[key]) is tuple:
        break_out = False
        for val in input_dict[key]:
          if type(val) is Empty:
            all_slots_filled = False
            break_out = True
          if (
            type(val) is tu.Tube and
            val.get_val() is None):
            all_slots_filled = False
            break_out = True
            break
        if break_out:
          break
    return all_slots_filled

  def _convert_tubes_to_vals(self, input_dict):
    """Pull out the values associated with the tubes connected to this tank's slots. Where the values of the dictionary are the values stored in the tube, rather than the Tube object itself.

    Parameters
    ----------
    input_dict : dict(
        keys - Slot keys. Must be the same as the attribute slot_keys.
        values - Tube, None or some valid input data type.
      )
      The inputs to the tank.

    Returns
    -------
    dict(
      keys - Slot keys. Must be the same as the attribute slot_keys.
      values - valid input data type.
    )
        The inputted dict with the Tube objects replaced by the 'val' attribute of the tube.

    """
    r_dict = {}
    r_dict.update(input_dict)
    for key in input_dict:
      if type(input_dict[key]) is tu.Tube:
        r_dict[key] = input_dict[key].get_val()

    return r_dict

  def _create_slots(self, slot_keys, waterwork):
    """Create all the slot objects for the tank and add to the waterwork directory.

    Parameters
    ----------
    slot_keys : list of str
      The tank's (operation's) argument keys. They define the names of the inputs to the tank.

    waterwork : Waterwork
      The waterwork that the part will be added to.

    """
    for key in slot_keys:
      slot = sl.Slot(self, key)
      self.slots[key] = slot

      waterwork.slots[slot.name] = slot
      waterwork.funnels[slot.name] = slot

  def _create_tubes(self, tube_keys, waterwork):
    """Create all the tube objects for the tank and add to the waterwork directory.

    Parameters
    ----------
    slot_keys : list of str
      The tank's (operation's) argument keys. They define the names of the inputs to the tank.

    waterwork : Waterwork
      The waterwork that the part will be added to.

    """
    for key in tube_keys:
      tube = tu.Tube(self, key)
      self.tubes[key] = tube

      waterwork.tubes[tube.name] = tube
      waterwork.taps[tube.name] = tube

  def _handle_iterable(self, iterable):
    found_tube = False
    for val in iterable:
      if type(val) is sl.Slot:
        raise ValueError("Cannot pass slot as argument to " + str(type(self)))

      if type(val) is tu.Tube or type(val) is Empty:
        found_tube = True
        break

    if not found_tube:
      return iterable

    import wtrwrks.tanks.tank_defs as td
    l_tubes, l_slots = td.tube_list(*iterable)

    for i, val in enumerate(iterable):
      if type(val) is not tu.Tube and type(val) is not Empty:
        l_slots['a' + str(i)].set_plug(val)

    return l_tubes['target']

  def _handle_merge(self):
    for key in self.slots:
      slot = self.slots[key]

      if type(slot.tube) is not Empty:
        slot.tube.downstream_tube = self.tubes['target']

  def _get_default_name(self, prefix=''):
    """Create the default name of the tank, of the form '<TankSubClass>_<num>'.

    Returns
    -------
    str
        The name of the tank.

    """
    num = 0
    cls_name = self.__class__.__name__

    # Start with the name being '<TankSubClass>_0'. If that is already taken,
    # keep increasing the number from 0 until an unused name is found.
    full_name = os.path.join(prefix, cls_name + '_' + str(num))
    while full_name in gl._default_waterwork.tanks:
      num += 1
      full_name = os.path.join(prefix, cls_name + '_' + str(num))

    return full_name

  def _join_tubes_to_slots(self, input_dict, waterwork):
    """Join the tubes incoming from other tanks to this tank's slots. If the slot was previously identified as a 'funnel', i.e. a slot not attached to other tube then take it out of the list of funnels. If the tube was previously identified as a tap, i.e. a tube that is not connected to another slot, then take it out of the list of taps.

    Parameters
    ----------
    input_dict : dict(
        keys - Slot keys. Must be the same as the attribute slot_keys.
        values - Tube, None or some valid input data type.
      )
      The inputs to the tank.
    waterwork : Waterwork
      The waterwork that the part will be added to.

    Returns
    -------
    type
        Description of returned object.

    """
    for key in input_dict:
      slot = self.slots[key]
      tube = input_dict[key]

      if type(tube) is not tu.Tube:
        continue

      # If the tube was already used for another tank, then it'll have to be
      # cloned.
      if type(tube.slot) is not Empty:

        # Save the slot in order to connect it to the clone tube later.
        other_slot = tube.slot
        tube.slot = empty

        import wtrwrks.tanks.tank_defs as td
        c_tubes, c_slots = td.clone(a=tube)

        # Join the other slot to the 'b' tube of the clone tank
        other_slot.tube = c_tubes['b']
        c_tubes['b'].slot = other_slot

        # Join this slot to the 'a' tube of the clone tank
        slot.tube = c_tubes['a']
        c_tubes['a'].slot = slot

        # Remove the newly created clone tupes from the taps, since they are
        # immediately connected to slots.
        del waterwork.taps[c_tubes['a'].name]
        del waterwork.taps[c_tubes['b'].name]

        # Remove the slots from the funnels, since the clone slot is now the
        # funnel.
        if slot.name in waterwork.funnels:
          del waterwork.funnels[slot.name]
          if slot.plug is not None:
            raise ValueError(str(slot) + ' was plugged but is no longer a funnel. Only funnels can have plugs.')
        if other_slot.name in waterwork.funnels:
          del waterwork.funnels[other_slot.name]
          if other_slot.plug is not None:
            raise ValueError(str(other_slot) + ' was plugged but is no longer a funnel. Only funnels can have plugs.')
      else:
        tube.slot = slot
        slot.tube = tube

      if type(tube) is tu.Tube:
        if tube.name in waterwork.taps:
          del waterwork.taps[tube.name]
        if slot.name in waterwork.funnels:
          del waterwork.funnels[slot.name]
        if slot.plug is not None:
          raise ValueError(str(slot) + ' was plugged but is no longer a funnel. Only funnels can have plugs.')
        if tube.plug is not None:
          raise ValueError(str(tube) + ' was plugged but is no longer a tap. Only taps can have plugs.')

  def _pour(self, *args, **kwargs):
    """Make sure pour function is set by subclass. The forward transformation of inputs."""
    raise ValueError("'_pour' method not defined for " + str(type(self)))

  def _pump(self, *args, **kwargs):
    """Make sure pump function is set by subclass. The backward transformation of outputs."""
    raise ValueError("'_pump' method not defined for " + str(type(self)))

  def _save_dict(self):
    save_dict = {}
    save_dict['func_name'] = self.func_name
    save_dict['name'] = self.name

    return save_dict

  def get_slot(self, key):
    """Retrieve a slot object from the tank identified by the key"""
    return self.slots[key]

  def get_slots(self):
    """Retrieve the dictionary of slot objects from the tank."""
    slots = {}
    slots.update(self.slots)
    return slots

  def get_tube(self, key):
    """Retrieve a tube object from the tank identified by the key"""
    return self.tubes[key]

  def get_tubes(self):
    """Retrieve the dictionary of tube objects from the tank."""
    tubes = {}
    tubes.update(self.tubes)
    return tubes

  def pour(self, **input_dict):
    """Execute the forward transformation of the input_dict inputted to the tank to get the dictionary of tube objects who's val's have been filled.

    Parameters
    ----------
    **input_dict : kwargs = {
        keys - Slot keys. Must be the same as the attribute slot_keys.
        values - valid input data types
      }
      The inputs to the tank.

    Returns
    -------
    kwargs = {
        keys - Tube keys. The same as the keys from attribute tube_keys.
        values - The data_types outputted by the tank.
      }
        All of the ouputs the tank gives in the 'pour' (i.e. forward) direction.

    """
    # Check that the inputs are valid
    if set(input_dict.keys()) != set(self.slot_keys):
      raise ValueError("Must pass " + str(input_dict.keys()) + " as arguments, got " + str(self.slot_keys))

    for key, val in input_dict.iteritems():
      if not self._slot_is_valid_type(key, val):
        raise TypeError("Got invalid type for (tank, slot): " + str((self.name, key)) + ". ")
    # Run the function defined by the subclass
    tube_dict = self._pour(**input_dict)

    # Set the vals
    for key in tube_dict:
      self.tubes[key].set_val(tube_dict[key])

    return tube_dict

  def pump(self, **kwargs):
    """Execute the backward transformation of the kwargs inputted to the tank to get the dictionary of slot objects who's val's have been filled.

    Parameters
    ----------
    **kwargs : kwargs = {
        keys - Tube keys. Must be the same as the keys from attribute tube_keys.
        values - valid data types
      }
      The inputs to the backward transformation of the tank.

    Returns
    -------
    kwargs = {
        keys - Slot keys. The same as the attribute slot_keys.
        values - The data_types outputted by the tank.
      }
        All of the ouputs the tank gives in the 'pump' (i.e. backward) direction.

    """
    # Check that the inputs are valid
    if set(kwargs.keys()) != set(self.tube_keys):
      raise ValueError("Must pass " + str(kwargs.keys()) + " as arguments, got " + str(self.tube_keys))

    for key, val in kwargs.iteritems():
      if not self._tube_is_valid_type(key, val):
        raise TypeError("Got invalid type for (tank, tube): " + str((self.name, key)) + ". ")
    # Run the function defined by the subclass
    slot_dict = self._pump(**kwargs)

    # Set the vals
    for key in slot_dict:
      self.slots[key].set_val(slot_dict[key])

    return slot_dict

  def get_slot_tanks(self):
    """Get a set of all the tanks that pour into this one.

    Returns
    -------
    set of Tanks
      The set of all tanks that feed into this one (in the pour direction).

    """
    tanks = set()
    for slot_key in self.slots:
      slot = self.slots[slot_key]
      if type(slot.tube) is not Empty:
        tanks.add(slot.tube.tank)
    return tanks

  def get_tube_tanks(self):
    """Get a set of all the tanks that pump into this one.

    Returns
    -------
    set of Tanks
      The set of all tanks that feed into this one (in the pump direction).

    """
    tanks = set()
    for tube_key in self.tubes:
      tube = self.tubes[tube_key]
      if type(tube.slot) is not Empty:
        tanks.add(tube.slot.tank)
    return tanks

  def get_pour_dependencies(self):
    """Get a set of tanks that need to be run before this tank is run (in the pour direction).

    Returns
    -------
    set of Tanks
      The set of tanks that need to be run before this one when executing a pour of a Waterwork.

    """
    tanks = [self]
    dependencies = set()

    while tanks:
      tank = tanks.pop()

      # Go through each of the slots of the current tank, to get to the
      # 'parent_tank', i.e. the tank that has outputs which are needed for the
      # current tanks inputs. Put the parent tank before the current one.
      for slot_name in tank.slots:

        slot = tank.slots[slot_name]

        # If the slot is not connected to any tube (i.e. is a funnel) continue.
        # print slot.name, slot.name in slot.waterwork.funnels
        if slot.name in slot.waterwork.funnels:
          continue

        parent_tank = slot.tube.tank

        tanks.append(parent_tank)
        dependencies.add(parent_tank)

    return dependencies

  def get_pump_dependencies(self):
    """Get a set of tanks that need to be run before this tank is run (in the pump direction).

    Returns
    -------
    set of Tanks
      The set of tanks that need to be run before this one when executing a pump of a Waterwork.

    """
    tanks = [self]
    dependencies = set()
    while tanks:
      tank = tanks.pop()

      # Go through each of the slots of the current tank, to get to the
      # 'parent_tank', i.e. the tank that has outputs which are needed for the
      # current tanks inputs. Put the parent tank before the current one.
      for tube_name in tank.tubes:
        tube = tank.tubes[tube_name]

        # If the tube is not connected to any slot (i.e. is a tap) continue.
        if tube.name in tube.waterwork.taps:
          continue

        parent_tank = tube.slot.tank

        tanks.append(parent_tank)
        dependencies.add(parent_tank)
    return dependencies

  def _slot_is_valid_type(self, key, val):
    return True

  def _tube_is_valid_type(self, key, val):
    return True
